from pathlib import Path
import numpy as np
import pandas as pd
from jahs_bench.tabular.clusterlib.prescheduler import JobConfig

required_evals_per_bucket = 10_000
workers_per_bucket_cap = required_evals_per_bucket
epochs_per_eval = 200
cpus_per_node = 48
total_num_nodes = 6400
available_num_cpus = cpus_per_node * total_num_nodes
max_evals_per_worker = 5

# Assuming linear memory scaling w.r.t N and W, quadratic scaling with Resolution
# For starters, it should suffice to ignore N and W and only scale according to Resolution
# TODO: Come back and revise this if something breaks.

fids = pd.MultiIndex.from_product([[1, 3, 5], [4, 8, 16], [0.25, 0.5, 1.0]], names=["N", "W", "Resolution"])
# fids = pd.MultiIndex.from_product([[1, 3, 5], [1, 2, 3], [8, 16, 32]], names=["N", "W", "Resolution"])
max_fid = (5, 16, 1.0)
max_memory_req = 6  # GB


def approximate_memory_reqs():
    req = pd.DataFrame(index=fids)
    req = req.assign(MemoryReq=max_memory_req)

    scales = 2 ** (2 * np.log2(req.index.get_level_values("Resolution").values)).reshape(-1, 1)
    req = req.mul(scales)
    return req


def final_calculations(runtimes_per_epoch):
    r = approximate_memory_reqs()
    # CPUs per node = memory_required / memory_per_cpu
    cpus_per_worker_per_node_per_bucket: pd.Series = (r / 1.5).clip(1).astype(int)["MemoryReq"]
    max_nodes_per_bucket = cpus_per_worker_per_node_per_bucket.rdiv(cpus_per_node).rdiv(workers_per_bucket_cap)
    max_nodes_per_bucket: pd.Series = max_nodes_per_bucket.clip(upper=total_num_nodes)
    workers_per_bucket = cpus_per_worker_per_node_per_bucket.rdiv(cpus_per_node) * max_nodes_per_bucket
    workers_per_bucket: pd.Series = workers_per_bucket.round()

    inv_time_ratios = runtimes_per_epoch.rdiv(runtimes_per_epoch.min())
    evals_per_worker = inv_time_ratios * max_evals_per_worker
    evals_per_worker = evals_per_worker.round().clip(lower=1).astype(int)
    workers_per_bucket /= evals_per_worker
    nodes_per_bucket = workers_per_bucket * cpus_per_worker_per_node_per_bucket / cpus_per_node
    nodes_per_bucket = nodes_per_bucket.round().astype(int)


    runtime_per_evaluation_per_bucket = runtimes_per_epoch * epochs_per_eval
    cpuh_required_per_bucket = runtime_per_evaluation_per_bucket * required_evals_per_bucket
    runtime_per_worker_per_bucket = cpuh_required_per_bucket / workers_per_bucket

    return cpus_per_worker_per_node_per_bucket, runtime_per_worker_per_bucket, evals_per_worker, nodes_per_bucket


def create_slurm_limits_dataframe(job_config: JobConfig, cpus_per_worker_per_node_per_bucket: pd.Series,
                                  cpuh_per_worker_per_bucket: pd.Series, evals_per_worker: pd.Series,
                                  nodes_per_bucket: pd.Series, worker_id_offset: int = 0,
                                  max_job_runtime: int = 24 * 60 * 60) -> pd.DataFrame:
    """ Use the fidelity-wise distribution of the resources required for some evaluation profile as well as a
    definition of the resources available per job as input in order to bundle up the total computation into the jobs
    that conform to the specified job resource profile. The bundled up jobs are described in the returned DataFrame as
    columns corresponding to resource requirements and indices corresponding to individual jobs and configurations,
    i.e. the index is a MultiIndex (job_id, config_id). It is assumed that all inputs have already been filtered
    appropriately such that there is no need to re-check the demand and supply of certain resources that are fixed
    across multiple bundled jobs, e.g. cpus per task. The resource requirements specified in the input should all be
    pandas Series with matching indices, such that each index value corresponds to a single fidelity bucket. The int
    'worker_id_offset' can be used to specify that these evaluations should be assumed to be continue a separate batch
    of evaluations, such that the last evalatuion in that batch had the worker id 'worker_id_offset - 1', or in other
    words 'worker_id_offset' evaluations have already been accounted for before this batch of evaluations.

    output dataframe format:
    index: MultiIndex, names = [job_id, config_id]
    columns: Index, values = [worker_id_start, worker_id_end, worker_id_offset,

    """

    cpus_per_worker_per_node_per_bucket: pd.Series = cpus_per_worker_per_node_per_bucket.astype(int)  # cpus per task
    evals_per_worker: pd.Series = evals_per_worker.astype(int)
    nodes_per_bucket: pd.Series = nodes_per_bucket.astype(int)

    ntasks_per_node = cpus_per_worker_per_node_per_bucket.rfloordiv(job_config.cpus_per_node)
    total_ntasks = (ntasks_per_node * nodes_per_bucket).sum()
    cpuh_required = cpuh_per_worker_per_bucket.sum() * total_ntasks




if __name__ == "__main__":
    cifar_total_num = 60_000
    colorectal_total_num = 5000
    ucmerced_total_num = 2100
    batch_size = 256

    runtimes_pth = Path("/home/archit/thesis/experiments/checkpointing/tables")
    cifar_runtimes_per_epoch: pd.Series = pd.read_pickle(runtimes_pth / "per_epoch_runtimes.pkl.gz")["mean"]

    ## CIFAR-10
    # outdir = Path("/home/archit/thesis/slurm_configs/cifar10")
    # runtimes_per_epoch = pd.read_pickle(runtimes_pth)["mean"]

    ## COLORECTAL_HISTOLOGY
    # outdir = Path("/home/archit/thesis/slurm_configs/colorectal_histology")
    # runtimes_pth = Path("/home/archit/thesis/slurm_configs/colorectal_histology/")
    # runtimes_per_epoch = pd.read_pickle(runtimes_pth / "approx_runtimes.pkl.gz")
    # memory_ratio = (5_000 / 50_000) * (128**2 / 32 ** 2)
    # max_memory_req = 5 * memory_ratio  # 5 GB base
    #
    # if "resolution" in runtimes_per_epoch.index.names:
    #     runtimes_per_epoch.index = runtimes_per_epoch.index.rename(runtimes_per_epoch.index.names[:2] + ["Resolution"])

    ## UC_MERCED
    outdir = Path("/home/archit/thesis/slurm_configs/uc_merced")
    # runtimes_pth = Path("/home/archit/thesis/slurm_configs/uc_merced/")
    img_size_ratio = 128**2 / 32 ** 2
    nbatches_ratio = (ucmerced_total_num // batch_size) / (cifar_total_num // batch_size)
    nimgs_ratio = ucmerced_total_num / cifar_total_num
    runtime_ratio = nbatches_ratio * img_size_ratio
    memory_ratio = nimgs_ratio * img_size_ratio

    runtimes_per_epoch = cifar_runtimes_per_epoch * runtime_ratio
    runtimes_per_epoch.to_pickle(outdir / "per_epoch_runtimes.pkl.gz")
    max_memory_req = 5 * memory_ratio  # 5 GB base

    if "resolution" in runtimes_per_epoch.index.names:
        runtimes_per_epoch.index = runtimes_per_epoch.index.rename(runtimes_per_epoch.index.names[:2] + ["Resolution"])



    # Map fidelity values to updated ones
    runtimes_per_epoch = runtimes_per_epoch[runtimes_per_epoch.index.get_level_values("N").isin([1, 3, 5])]
    map_W = {1: 4, 2: 8, 3: 16}
    map_Res = {8: 0.25, 16: 0.5, 32: 1.0}
    runtimes_per_epoch.index = runtimes_per_epoch.index.map(lambda x: (x[0], map_W[x[1]], map_Res[x[2]]))

    cpus_per_worker_per_node_per_bucket, runtime_per_worker_per_bucket, evals_per_worker, nodes_per_bucket = \
        final_calculations(runtimes_per_epoch)

    # These are all over the place. This entire script should be phased out in favor of a better profiler that
    # distributes the entire workload across multiple jobs of a given, fixed resource size instead of trying to infer
    # optimal resource allocations from the estimates. The difference in name between the variables and their pickle
    # filenames has been intentionally left in in order to indicate differences in behaviour and interpretation going
    # forward.
    cpus_per_worker_per_node_per_bucket.to_pickle(outdir / "cpus_per_worker_per_node_per_bucket.pkl.gz")
    runtime_per_worker_per_bucket.to_pickle(outdir / "cpuh_per_worker_per_bucket.pkl.gz")
    evals_per_worker.to_pickle(outdir / "evals_per_worker.pkl.gz")
    nodes_per_bucket.to_pickle(outdir / "nodes_per_bucket.pkl.gz")
    pass



"""
This script is intended to verify that a given stage of the benchmark is reproducible.
The underlying idea is that for each sampled configuration, it should be possible to
reconstruct the network using the saved model configuration, load the second to last
checkpoint, resume training for a small number of epochs and cross-check the generated
metrics against the saved metric values from the previous run (i.e. the latest metrics
DataFrame). A subset of all recorded metrics are deterministic and the rest are
stochastic. This script should report the observed difference in stochastic metric values
as well as catch instances where the deterministic metric values differed.
"""

import argparse
import logging
from pathlib import Path

from jahs_bench.tabular.lib.core import constants

duration_metrics = [m for m in constants.standard_model_dataset_metrics
                    if "duration" in metric]
data_splits = ["train", "valid", "test"]
stochastic_metrics = [(s, m) for s in data_splits for m in duration_metrics] + \
                     [("train", m) for m in constants.extra_model_training_metrics] + \
                     [("diagnostic", m) for m in ["latency", "runtime"]]

deterministic_metrics = [(s, m) for s in data_splits for m in ["loss", "acc"]] + \
                        [("diagnostic", m) for m in ["FLOPS", "size_MB"]]


_log = logging.getLogger(__name__)


def argument_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("--rootdir", type=Path, default=Path().cwd(),
                        help="Path to the root directory where all the tasks' output was stored. Task-specific "
                             "sub-directories will be created here if needed. Note that this is NOT the same as the "
                             "'basedir' of a DirectoryTree object as multiple DirectoryTree objects will be created "
                             "as and when needed.")
    parser.add_argument("--tmp_rootdir", type=Path, default="/tmp/nashpo_benchmarks/rootdir",
                        help="A directory which will be used to store temporary data, "
                             "essentially creating a temporary partial copy of the metrics and checkpoints of the original ")
    parser.add_argument("--workerid", type=int,
                        help="An offset from 0 for this worker within the current job's allocation of workers. This "
                             "does NOT correspond to a single 'taskid' but rather the portfolio of configurations that "
                             "this worker will evaluate.")
    parser.add_argument("--workerid-offset", type=int, default=0,
                        help="An additional fixed offset from 0 for this worker that, combined with the value of "
                             "'--worker', gives the overall worker ID in the context of a job that has been split into "
                             "multiple smaller parts.")
    parser.add_argument("--nworkers", type=int,
                        help="The total number of workers that are expected to concurrently perform data verification "
                             "in this root directory. This is used to coordinate the workers such that there is no "
                             "overlap in their portfolios.")
    parser.add_argument("--datadir", type=Path,
                        help="The directory where all datasets are expected to be stored.")
    parser.add_argument("--configs", type=Path,
                        help="Path to a pickled Pandas DataFrame (*.pkl.gz) that contains the full configuration of "
                             "each known model that should be verified. The index of each such configuration should be "
                             "the unique tuple (taskid, model_idx), where the corresponding values will be used to "
                             "construct corresponding directory trees.")
    parser.add_argument("--nepochs", type=int, default=None,
                        help="Only used in the 'verify' mode of operation. The number of epochs of data that should be "
                             "generated and verified using each loaded checkpoint. This implies that, for example, in "
                             "a model with 5 pairs of checkpoints and metric logs that have already been verified to "
                             "be in sync, the first 4 checkpoints will each be loaded up, the model trained for "
                             "'nepochs' epochs, and the corresponding entries in the last 4 metrics logs will be "
                             "cross-checked.")
    parser.add_argument("--budget", type=int,
                        help="Time budget for each worker in seconds. This is used to approximate whether or not a "
                             "worker has enough time to perform one more operation without any risk to any data.")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")

    return parser



import json
from copy import deepcopy
from pathlib import Path
from functools import partial
from typing import Callable, Sequence, Optional

import numpy as np
import torch
from jahs_bench.tabular.lib.naslib.utils.utils import AttrDict, Cutout
from torchvision import datasets as dset, transforms as transforms

from jahs_bench.tabular.lib.core import constants as constants
from jahs_bench.tabular.lib.core.aug_lib import TrivialAugment

from icgen.vision_dataset import ICVisionDataset

"""
Adapted in large part from the original NASBench-201 code repository at
https://github.com/D-X-Y/AutoDL-Projects/tree/bc4c4692589e8ee7d6bab02603e69f8e5bd05edc
"""


def get_dataloaders(dataset: constants.Datasets, batch_size: int, cutout: int = -1, split: bool = True,
                    resolution: int = 1., trivial_augment=False, datadir: Path = None):

    if not isinstance(dataset, constants.Datasets):
        raise TypeError(f"A dataset name should be an instance of {constants.Datasets}, was given {type(dataset)}.")

    if resolution <= 0. or resolution > 1.:
        raise ValueError(f"Invalid image resolution scaling: {resolution}. Should be a value between 0. and 1.")

    dataset_fns = {**{
        constants.Datasets.cifar10: dset.CIFAR10,
        # constants.Datasets.fashionMNIST: dset.FashionMNIST,
    }, **{d: partial(load_icgen_dataset, name=d.name) for d in constants.icgen_datasets}}

    name_str, image_size, nchannels, nclasses, mean, std, train_size, test_size = dataset.value
    crop_size = image_size
    padding_size = max(0, int(4 * resolution))
    image_size = int(resolution * image_size)

    datadir = get_default_datadir() if datadir is None else datadir
    if dataset in constants.icgen_datasets:
        # Ugly hack. This data would automatically be read by ICVisionDataset but we need the mean/std values before
        # that object is initialized.
        crop_size = image_size
        datadir = datadir / "downsampled" / str(image_size)
        with open(datadir / dataset.name / "info.json") as fp:
            meta = json.load(fp)
        mean = [m / 255 for m in meta["mean_pixel_value_per_channel"]]
        std = [m / 255 for m in meta["mean_std_pixel_value_per_channel"]]

    if dataset not in dataset_fns:
        raise NotImplementedError(f"Pre-processing for dataset {name_str} has not yet been implemented.")

    # Data Augmentation
    lists = [TrivialAugmentTransform()] if trivial_augment else []
    lists += [transforms.RandomHorizontalFlip(), transforms.RandomCrop(crop_size, padding=padding_size)]
    # ICGen datasets have been resized already so we avoid downsampling twice.
    lists += [transforms.Resize(image_size)] if resolution < 1. and dataset not in constants.icgen_datasets else []
    lists += [transforms.ToTensor(), transforms.Normalize(mean, std)]

    if cutout > 0 and not trivial_augment:  # Trivial Augment already contains Cutout
        lists += [Cutout(cutout)]

    train_transform = transforms.Compose(lists)
    test_transform = transforms.Compose(
        ([transforms.Resize(image_size)] if resolution < 1. and dataset not in constants.icgen_datasets else [])
        + [transforms.ToTensor(), transforms.Normalize(mean, std)]
    )
    min_shape = (1, nchannels, image_size, image_size)
    train_data = dataset_fns[dataset](root=datadir, train=True, transform=train_transform, download=False)
    test_data = dataset_fns[dataset](root=datadir, train=False, transform=test_transform, download=False)

    assert len(train_data) == train_size, f"Invalid dataset configuration, expected {train_size} images, got " \
                                          f"{len(train_data)} for dataset {name_str}."
    assert len(test_data) == test_size, f"Invalid dataset configuration, expected {test_size} images, got " \
                                        f"{len(test_data)} for dataset {name_str}."

    test_loader = torch.utils.data.DataLoader(
        test_data,
        batch_size=batch_size,
        shuffle=False,
        num_workers=0,
        pin_memory=True,
    )

    test_transform = test_data.transform

    loaders = {"test": test_loader}
    if split:
        ## Split original training data into a training and a validation set, use test data as a test set
        splitdir = datadir / dataset.name if dataset in constants.icgen_datasets else datadir
        split_info = load_splits(path=splitdir / f"{dataset.name}-validation-split.json")
        assert len(train_data) == len(split_info.train) + len(split_info.valid), \
            f"invalid length : {len(train_data)} vs {len(split_info.train)} + {len(split_info.valid)}"
        valid_data = deepcopy(train_data)
        valid_data.transform = test_transform
        # data loader
        train_loader = torch.utils.data.DataLoader(
            train_data,
            batch_size=batch_size,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(split_info.train),
            num_workers=0,
            pin_memory=True,
        )
        valid_loader = torch.utils.data.DataLoader(
            valid_data,
            batch_size=batch_size,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(split_info.valid),
            num_workers=0,
            pin_memory=True,
        )
        loaders["train"] = train_loader
        loaders["valid"] = valid_loader
    else:
        # data loader
        train_loader = torch.utils.data.DataLoader(
            train_data,
            batch_size=batch_size,
            shuffle=True,
            num_workers=0,
            pin_memory=True,
        )
        loaders["train"] = train_loader

    return loaders, min_shape


def load_splits(path: Path):
    assert path.exists(), "Can not find {:}".format(path)
    # Reading data back
    with open(path, "r") as f:
        data = json.load(f)
    splits = {k: np.array(v[1], dtype=int) for k, v in data.items()}
    del data
    return AttrDict(splits)


def get_default_datadir() -> Path:
    return Path(__file__).parent.parent.parent / "data"


class TrivialAugmentTransform(torch.nn.Module):
    def __init__(self):
        self._apply_op = TrivialAugment()
        super(TrivialAugmentTransform, self).__init__()

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return self._apply_op(img)


def load_icgen_dataset(name: str, root: Path, train: bool = True,
                       transform: Optional[Sequence[Callable]] = None,
                       target_transform: Optional[Sequence[Callable]] = None,
                       download: bool = False) -> ICVisionDataset:
    """ Load an ICVisionDataset. This function serves as a wrapper around the underlying ICVisionDataset initializer
    in order to provide an interface compatible with calls to most Torchvision.Dataset classes. The 'download'
    parameter has been provided for compatibility only, the actual dataset should be downloaded and prepared in advance.
    """

    if download:
        raise UserWarning("The parameter 'download' has been provided for compatibility purposes only. The actual "
              "dataset should be downloaded and prepared in advance using ICGen.")

    dataset = ICVisionDataset(dataset=name, root=root,
                              split="train" if train else "test",
                              transform=transform, target_transform=target_transform)
    return dataset



"""
This library contains various functions that are intended to help generate a pre-computed schedule for distributing
a large number of model evaluations across a set number of jobs with certain properties. The intention is to generate
a unique portfolio file for each unique and independent worker in a distributed computing environment.
"""

import itertools
import logging
import math
import pandas as pd
from pathlib import Path
from typing import Optional, Sequence, Iterator, Tuple, Union, Dict

from jahs_bench.tabular.lib.core import constants, utils

_log = logging.getLogger(__name__)
fidelity_params = constants.fidelity_params
fidelity_types = constants.fidelity_types
logdir_name = "logs"


class JobConfig(object):
    """ A container for various properties of the cluster that will be used to help compute the work schedule. """

    cpus_per_worker: int
    cpus_per_node: int
    nodes_per_job: int
    timelimit: float

    def __init__(self, cpus_per_worker=1, cpus_per_node=48, nodes_per_job=3072, timelimit=2 * 24 * 3600):
        self.cpus_per_worker = cpus_per_worker
        self.cpus_per_node = cpus_per_node
        self.nodes_per_job = nodes_per_job
        self.timelimit = timelimit

    @property
    def workers_per_node(self) -> int:
        return self.cpus_per_node // self.cpus_per_worker

    @property
    def workers_per_job(self) -> int:
        return self.nodes_per_job * self.workers_per_node

    @property
    def template_kwargs(self) -> dict:
        """ A dict of keyword arguments that should provide most of the details needed to fill in a job template. """
        return {
            "cpus_per_task": self.cpus_per_worker,
            # "cpus_per_node": self.cpus_per_node,
            "nodes_per_job": self.nodes_per_job,
            "tasks_per_node": self.workers_per_node,
            "timelimit": self.timestr(self.timelimit),
        }

    @staticmethod
    def timestr(secs: float) -> str:
        """ Generates a human-readable and SLURM-compatible string representing the given duration in seconds as
        DD-HH:MM, where DD is the number of days, HH is the number of hours and MM is the number of minutes.
        Seconds-level precision is omitted since it is unlikely to be very useful. """

        minutes = math.ceil(secs / 60)
        hours = minutes // 60
        minutes %= 60
        days = hours // 24
        hours %= 24
        return f"{days}-{hours:02}:{minutes:02}"


class WorkerConfig(object):
    """ Container of properties for each worker's work allocation. """

    worker_id: int
    portfolio_dir: Optional[Path]
    portfolio: Optional[pd.Series]

    def __init__(self, worker_id: int, portfolio_dir: Optional[Path] = None, portfolio: Optional[pd.DataFrame] = None):
        self.worker_id = worker_id
        self.portfolio_dir = portfolio_dir
        self.portfolio = portfolio

    @property
    def portolio_file(self) -> Path:
        assert self.portfolio_dir is not None, "No portfolio directory specified, cannot generate portfolio file name."
        return self.portfolio_dir / f"portfolio_{self.worker_id}.pkl.gz"

    def load_portfolio(self):
        if self.portfolio is not None:
            _log.warning(f"Existing portfolio of worker {self.worker_id} will be overwritten by portfolio read from "
                         f"{self.portolio_file}")
        self.portfolio: pd.Series = pd.read_pickle(self.portolio_file)

    def iterate_portfolio(self, start: int = 0, stop: int = None, step: int = 1, rootdir: Optional[Path] = None) -> \
            Iterator[Tuple[int, int, Path]]:
        """ Returns an iterator over this worker's portfolio, consisting of a list of 3-tuples, each consisting of
        an integer task ID, an integer model index, and a path to the base directory where all the metrics and
        checkpoints of this configuration's previous evaluations are stored. The optional Path-like 'rootdir' can be
        specified to read all the saved base directory paths to be treated as relative to the specified 'rootdir'. """

        index = self.portfolio.index[start:stop:step]
        pths = self.portfolio[index]
        if rootdir is None:
            pths = [Path().cwd() / p for p in pths]
        else:
            pths = [(rootdir / p).resolve() for p in pths]

        return iter([(*c, p) for c, p in zip(index, pths)])

    def __hash__(self):
        return hash(self.worker_id)


WorkAssignment = Dict[WorkerConfig, Tuple[float, Sequence[Tuple[int, int]]]]


class JobAllocation:
    def __init__(self, config: JobConfig, worker_id_offset: int = 0, assignment: Optional[WorkAssignment] = None,
                 dynamic_nnodes: bool = True, dynamic_timelimit: bool = True):
        """ A container for a single job's work assignment. Used primarily for bundling together workers operating on
        the same resource assignment (JobConfig) that have been assigned some portfolio into a single job file. """
        self.config = JobConfig(cpus_per_worker=config.cpus_per_worker, cpus_per_node=config.cpus_per_node,
                                nodes_per_job=config.nodes_per_job, timelimit=config.timelimit)
        self.worker_id_offset = worker_id_offset
        self.assignment = assignment
        self.dynamic_nnodes = dynamic_nnodes
        self.dynamic_timelimit = dynamic_timelimit

    @property
    def assignment(self) -> Dict[WorkerConfig, Tuple[float, Sequence[Tuple[int, int]]]]:
        """ """
        return self._assignment

    @assignment.setter
    def assignment(self, value: WorkAssignment):
        if value is None:
            # Initialize with an empty work assignment
            self.workers: Sequence[WorkerConfig] = [WorkerConfig(worker_id=self.worker_id_offset + i)
                                                    for i in range(self.config.workers_per_job)]
            self._assignment = {w: [0., []] for w in self.workers}
        else:
            assert isinstance(value, dict), f"The work assignment of a JobAllocation must be a dictionary of the " \
                                            f"form {WorkAssignment}."
            self.workers: Sequence[WorkerConfig] = list(value.keys())
            self._assignment = value

    @property
    def cpuh_demand(self) -> float:
        """ Total requested CPUh """
        return self.config.workers_per_job * self.config.cpus_per_worker * self.config.timelimit / 3600

    @property
    def cpuh_utilization(self) -> float:
        """ Total active runtime of all workers. """
        return sum([allocation[0] for allocation in self.assignment.values()]) * self.config.cpus_per_worker / 3600

    def schedule_work(self, profile: pd.DataFrame, runtime_estimates: pd.Series) -> pd.DataFrame:
        """ Attempts to greedily fill up all available workers within this job allocation and returns the part of the
        input profile that could not be fit within this job allocation. """

        assert profile.index.isin(runtime_estimates.index).all(), "Mismatch between the index of the given runtime " \
                                                                  "estimates and the profile."
        job_config = self.config
        workers = self.workers
        assignment = self.assignment
        worker_id_cycle = itertools.cycle(workers)
        curr_worker = next(worker_id_cycle)

        def find_next_eligible_worker(required_runtime: float) -> Optional[WorkerConfig]:
            """ Greedily fill up each worker's budget before moving on to the next worker. """
            nonlocal curr_worker
            counter = itertools.count()
            while next(counter) < job_config.workers_per_job:
                available_runtime = job_config.timelimit - assignment[curr_worker][0]
                if available_runtime >= required_runtime:
                    return curr_worker
                else:
                    curr_worker = next(worker_id_cycle)
            return None

        remainder = []
        for conf in profile.index:
            runtime = runtime_estimates[conf]
            assigned_worker = find_next_eligible_worker(required_runtime=runtime)
            if assigned_worker is None:
                remainder.append(conf)
                continue

            assignment[assigned_worker][0] += runtime
            assignment[assigned_worker][1].append(conf)

        for worker in workers:
            basedirs: pd.Series = profile.loc[assignment[worker][1], ("job_config", "basedir")].rename("basedir")
            basedirs.sort_index(axis=0, inplace=True)
            worker.portfolio = basedirs

        # final_assignment = {k: tuple(v) for k, v in assignment.items()}
        remainder = profile.loc[remainder, :]

        return remainder

    def optimize(self):
        """ If the flags 'dynamic_timelimit' and 'dynamic_nnodes' are True, optimize the job's resource demands w.r.t.
        the allocated work so as to maximize the CPUh utilization. """

        _log.debug(f"Current CPUh utilization: {self.cpuh_utilization * 100 / self.cpuh_demand:.2f}%")
        job_config = self.config

        if self.dynamic_nnodes:
            # If any workers don't have any work assigned to them, don't request those nodes in the first place.
            empty_workers = []
            for worker, (runtime, configs) in self.assignment.items():
                if runtime == 0.:
                    empty_workers.append(worker)

            for worker in empty_workers:
                self.assignment.pop(worker)

            self.workers = list(self.assignment.keys())

            empty_cpus = len(empty_workers) * job_config.cpus_per_worker
            empty_nodes = empty_cpus // job_config.cpus_per_node
            if empty_nodes > 0:
                job_config.nodes_per_job -= empty_nodes
                _log.debug(f"Reduced number of nodes required by {empty_nodes}.New CPUh utilization is at "
                           f"{self.cpuh_utilization * 100 / self.cpuh_demand:.2f}%")

        if self.dynamic_timelimit:
            # Clip the job runtime to the maximum required by any worker.
            max_runtime = max([allocation[0] for allocation in self.assignment.values()])
            job_config.timelimit = max_runtime
            _log.debug(f"Re-adjusted the job length to (DD-HH:MM): {JobConfig.timestr(job_config.timelimit)}. New CPUh "
                       f"utilization is at {self.cpuh_utilization * 100 / self.cpuh_demand:.2f}%")

    def save_worker_portfolios(self, jobid: int, portfolio_dir: Path) -> Path:
        """ Saves the respective portfolios for each worker in the given directory 'portfolio_dir'. The directory
        is created if it doesn't already exist. """

        portfolio_dir.mkdir(exist_ok=True, parents=False)
        nworkers = len(self.workers)
        portfolios = {w.worker_id: w.portfolio for w in self.workers}
        job_portfolio = pd.concat(portfolios, axis=0)
        # TODO: Save index level name in constants
        job_portfolio.index = job_portfolio.index.set_names("worker_id", level=0)
        portfolio_fn = portfolio_dir / f"portfolio_{jobid}.pkl.gz"
        job_portfolio.to_pickle(portfolio_fn)
        return portfolio_fn
        # for i, worker in enumerate(self.workers, start=1):
        #     worker.portfolio_dir = portfolio_dir
        #     worker.portfolio.to_pickle(worker.portolio_file)
        #
        #     if (i % (5 * max(1, nworkers // 100))) == 0:
        #         _log.info(f"Saved {i / nworkers} portfolios ({i * 100 / nworkers:.2f}%).")


def fidelity_basedir_map(c: Union[pd.Series, dict]):
    """ Given a Series or mapping from fidelity parameter names to values, generate a fidelity-specific base directory
    name. """

    global fidelity_params, fidelity_types
    return "-".join([f"{p}-{fidelity_types[p](c[i])}" for i, p in enumerate(fidelity_params)]) + "/tasks"


# noinspection PyPep8
def allocate_work(job_config: JobConfig, profile: pd.DataFrame, epochs: int, cpuh_utilization_cutoff: float = 0.75,
                  dynamic_timelimit: bool = True, dynamic_nnodes: bool = True, remainder_pth: Path = None,
                  worker_id_offset: int = 0, squish_configs: bool = False) -> Sequence[JobAllocation]:
    """
    Given a job configuration with estimates of each model's required runtime per epoch, generates a work schedule in
    the form of a list of WorkerConfig objects that have been allocated their respective portfolios.
    'runtime_estimates' should be a pandas DataFrame with a MultiIndex index containing model IDs as defined by the
    unique tuple (<taskid>, <model_idx>), and MultiIndex columns with values [('model_config', <conf>)] and
    [('runtime', 'required')], where <conf> represents all the names of the parameters in the config.
    Scheduling is performed using a not very efficient O(m.n) greedy solver, where m is the number of configurations
    to be scheduled and n is the number of workers available.
    """

    per_epoch_runtimes: pd.Series = profile.required.runtime

    if ("status", "nepochs") in profile.columns:
        remaining_epochs: pd.Series = epochs - profile.status.nepochs
    else:
        remaining_epochs: int = epochs

    runtime_estimates: pd.Series = per_epoch_runtimes * remaining_epochs
    runtime_estimates = runtime_estimates.sort_values(axis=0, ascending=False)

    sel = runtime_estimates > 0.
    if sel.index.size != profile.index.size:
        _log.info(f"Filtering out {profile.index.size - sel.index.size} negative values for required runtime.")

    runtime_estimates = runtime_estimates[sel]
    profile = profile.loc[runtime_estimates.index]

    sel = runtime_estimates.notna()
    if sel.index.size != profile.index.size:
        _log.info(f"Filtering out {profile.index.size - sel.index.size} NaN values for required runtime.")

    runtime_estimates = runtime_estimates[sel]
    profile = profile.loc[runtime_estimates.index]

    if squish_configs:
        squishable = runtime_estimates[runtime_estimates > job_config.timelimit]
        _log.info(f"Attempting to squish the {squishable.index.size} extra long configs into the given job "
                  f"resource allocation.")
        runtime_estimates[squishable.index] = job_config.timelimit

    total_required_budget = runtime_estimates.sum()

    _log.info(f"Estimated total CPUh requirement: "
              f"{total_required_budget * job_config.cpus_per_worker / 3600:<,.2f}")

    _log.info(f"Generating work schedule for {runtime_estimates.size} configurations spread across "
              f"{job_config.workers_per_job} workers per job.")

    # workers = [WorkerConfig(worker_id=i) for i in range(available_num_workers)]
    # work = {w: [job_config.timelimit, []] for w in workers}
    #
    # worker_id_cycle = itertools.cycle(workers)
    #
    # def find_next_eligible_worker(required_runtime: float) -> Optional[WorkerConfig]:
    #     counter = itertools.count()
    #     while next(counter) < available_num_workers:
    #         curr_worker = next(worker_id_cycle)
    #         available_runtime = work[curr_worker][0]
    #         if available_runtime >= required_runtime:
    #             return curr_worker
    #     return None
    #
    # for conf in runtime_estimates.index:
    #     runtime = runtime_estimates[conf]
    #     assigned_worker = find_next_eligible_worker(required_runtime=runtime)
    #     if assigned_worker is None:
    #         _log.info(f"Found no available work slots for config id {conf}, required runtime - "
    #                   f"{JobConfig.timestr(runtime)}")
    #         continue
    #
    #     work[assigned_worker][0] -= runtime
    #     work[assigned_worker][1].append(conf)
    #
    # for worker in workers:
    #     basedirs: pd.Series = profile.loc[work[worker][1], ("job_config", "basedir")].rename("basedir")
    #     basedirs.sort_index(axis=0, inplace=True)
    #     worker.portfolio = basedirs

    current_profile = profile
    jobs = []
    while True:
        job_allocation = JobAllocation(config=job_config, worker_id_offset=worker_id_offset,
                                       dynamic_nnodes=dynamic_nnodes, dynamic_timelimit=dynamic_timelimit)
        remainder = job_allocation.schedule_work(profile=current_profile, runtime_estimates=runtime_estimates)

        if remainder.index.size == current_profile.index.size:
            # The job timelimit is too small for these configs, which is why no more work can be assigned

            _log.warning(f"The given job timelimit {JobConfig.timestr(job_config.timelimit)} (DD-HH:MM) is too small "
                         f"for {remainder.index.size} configs with an average runtime requirement of "
                         f"{JobConfig.timestr(runtime_estimates[remainder.index].mean())} (DD-HH:MM). The remaining "
                         f"configs were distributed across {len(jobs)} jobs.")

            if remainder_pth is not None:
                _log.info(f"Saving {remainder.index.size} leftover configs at {remainder_pth}")
                remainder.to_pickle(remainder_pth)

            break
        else:
            # A job allocation has been successful. Add it to the list of jobs.
            job_allocation.optimize()
            jobs.append(job_allocation)
            worker_id_offset += job_allocation.config.workers_per_job
            if remainder.index.size == 0:
                _log.info(
                    f"Distributed the workload across {len(jobs)} jobs.")
                break
            else:
                current_profile = remainder

    utilization = [job.cpuh_utilization for job in jobs]
    demand = [job.cpuh_demand for job in jobs]
    avg_utilization = sum(utilization) / sum(demand)
    utilization = [u / d for u, d in zip(utilization, demand)]
    underutilized = avg_utilization  < cpuh_utilization_cutoff

    nworkers = sum([j.config.workers_per_job for j in jobs])

    if underutilized:
        _log.warning(f"The current job setup may not utilize all available workers very well. The current setup "
                     f"demands {sum(demand):<,.2f} CPUh and has a predicted approximate CPUh utilization of "
                     f"{avg_utilization * 100:.2f}%. Individual jobs have CPUh utilization in the range of "
                     f"{min(utilization) * 100:.2f}% to {max(utilization) * 100:.2f}%, spread over {nworkers} workers. "
                     f"To continue adding workers to this job allocation, use the worker ID offset {worker_id_offset}.")
    else:
        _log.info(f"The job setup has an overall CPUh utilization factor of {avg_utilization * 100:.2f}%. Individual "
                  f"jobs have CPUh utilization in the range of {min(utilization) * 100:.2f}% to "
                  f"{max(utilization) * 100:.2f}%, spread over {nworkers} workers. To continue adding workers to this "
                  f"job allocation, use the worker ID offset {worker_id_offset}.")

    return jobs


def generate_model_ids(specs: pd.DataFrame):
    """ Generate all expected model ids i.e. (taskid, model_idx), given the specifications in 'specs'. 'specs' should
    be a dataframe with the fidelity values as index and the columns [ntasks, nsamples] as columns. For each fidelity
    value i.e. row, it is assumed that 'ntasks' different task ids with 'nsamples' model_idx values each are needed.
    The fidelity values are then converted into partial model configs for the newly generated model ids. Thus, the
    returned dataframe will have an index corresponding to every (taskid, model_idx) expected to be generated by the
    input specifications. The columns will instead contain the fidelity values corresponding to each model id. """

    global fidelity_params

    assert "ntasks" in specs.columns, "The 'specs' dataframe should contain the column 'ntasks'."
    assert "nsamples" in specs.columns, "The 'specs' dataframe should contain the column 'nsamples'."

    for f in fidelity_params:
        assert f in specs.index.names, f"The fidelity parameter {f} is missing from the index of the specifications " \
                                       f"DataFrame"

    specs = specs.reorder_levels(fidelity_params, axis=0)
    partial_portfolios = []
    taskid_offset = 0

    for f in specs.index:
        ntasks = specs.loc[f]["ntasks"]
        nsamples = specs.loc[f]["nsamples"]
        index = pd.MultiIndex.from_product(
            [list(range(taskid_offset, taskid_offset + ntasks)), list(range(1, 1 + nsamples))],
            names=[constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value]
        )
        df = pd.DataFrame({n: f[i] for i, n in enumerate(fidelity_params)}, index=index, columns=fidelity_params)
        partial_portfolios.append(df)
        taskid_offset += ntasks

    full_portfolio = pd.concat(partial_portfolios, axis=0)
    return full_portfolio


def generate_basedirs(specs: pd.DataFrame) -> pd.Series:
    """ Given a dataframe such that the index values are unique model ids - (taskid, model_idx) tuples - and the
    columns are fidelity values corresponding to that model id, returns a Series "basedir" that contains a string which
    can be suffixed to a common root directory in order to generate a fidelity-wise base directory for a DirectoryTree
    object. """

    assert constants.MetricDFIndexLevels.taskid.value in specs.index.names, \
        f"The input 'specs' DataFrame must contain the index level {constants.MetricDFIndexLevels.taskid.value}"
    assert constants.MetricDFIndexLevels.modelid.value in specs.index.names, \
        f"The input 'specs' DataFrame must contain the index level {constants.MetricDFIndexLevels.modelid.value}"

    # def mapper(fidelity: pd.Series) -> str:
    #     return "-".join([f"{n}-{fidelity[n]}" for n in fidelity.index])

    return specs.apply(fidelity_basedir_map, axis=1).rename("basedir")


def prepare_directory_structure(basedirs: pd.Series, rootdir: Path):
    """ Given a Series 'basedirs' as input, with index consisting of tuples (taskid, model_idx) and directory names
    (strings) as the data values, and a path to a root directory, creates a large directory tree such that all future
    benchmarking operations in this root directory for the model configs identified by the input Series' index can
    assume that all relevant directories already exist. """

    _log.info("Preparing benchmark data directory structure.")

    assert rootdir.exists(), f"The root directory {rootdir} must be created before the prepare function can be called."
    index_levels = [constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value]
    logdir = rootdir / logdir_name
    logdir.mkdir(exist_ok=True, parents=False)

    for level in index_levels:
        assert level in basedirs.index.names, f"The input Series must have {level} as an index level."

    basedirs = basedirs.reorder_levels(index_levels)
    nconfigs = basedirs.size

    for i, (taskid, model_idx) in enumerate(basedirs.index, start=1):
        base = rootdir / basedirs[taskid, model_idx]
        base.mkdir(exist_ok=True, parents=True)
        try:
            tree = utils.DirectoryTree(basedir=base, read_only=False)
        except FileExistsError:
            pass

        try:
            tree.taskid = taskid
        except FileExistsError:
            pass

        try:
            tree.model_idx = model_idx
        except FileExistsError:
            pass

        if i % (5 * max(nconfigs // 100, 1)) == 0:
            # Report progress after iterating over approximately every 5 percent of the configs
            _log.info(f"Finished:\t{i * 100 / nconfigs:=6.2f}%\t({i}/{nconfigs}).")

    _log.info("Finished creating the directory structure.")


# TODO: Overhaul the entire profile generation script to instead work on the concept of accepting only expected mean
#  runtime per epoch, cpus per task/fidelity, required number of evaluations per fidelity, required number of epochs
#  and the available resources per job in order to distribute the workload across the workers. ntasks per fidelity
#  should also be a user-specified fixed value whose purpose is not to ensure that workers don't clash but to control
#  the spread of the directory tree as well as the number of parallel bit streams.
def prepare_full_sampling_profile(cpus_per_task_per_bucket: pd.Series,
                                  runtime_per_task_per_bucket: pd.Series, evals_per_task: pd.Series,
                                  nodes_per_bucket: pd.Series):
    """ Given fidelity-wise estimates of various resource requirements, generate a complete sampling profile containing
    the resource requirements of each individual model config - identified by the tuple (taskid, model_idx) - along
    with their respective base directories and partial configs (fidelity values) as a single DataFrame.

    Final DataFrame structure:
    index - taskid, model_idx
    columns - (required, runtime), (required, cpus), [(model_config, <fidelity parameters>)], (job_config, basedir)
    """

    global fidelity_params

    cpus_per_task_per_bucket: pd.Series = cpus_per_task_per_bucket.astype(int)
    evals_per_task: pd.Series = evals_per_task.astype(int)
    nodes_per_bucket: pd.Series = nodes_per_bucket.astype(int)

    # TODO: Fix this. The input should be workers per bucket - nodes per bucket, cpus per bucket, etc. can easily be
    #  inferred from the current resource availability and should be inferred on an ad-hoc basis

    # 48 = cpus per node, this has been fixed because this is the underlying assumption for the calculations behind
    # the estimates.
    ntasks_per_node = cpus_per_task_per_bucket.rfloordiv(48)
    total_ntasks = ntasks_per_node * nodes_per_bucket
    model_ids = generate_model_ids(
        specs=pd.concat([total_ntasks.rename("ntasks"), evals_per_task.rename("nsamples")], axis=1)
    )
    basedirs = generate_basedirs(specs=model_ids)

    # This is a change in philosophy for how the jobs should be built-up going forward as opposed to what was done thus
    # far. In a future version, this hack for converting runtimes back to runtimes per epoch won't be necessary.
    # noinspection PyArgumentList
    runtime_per_evaluation_per_epoch = runtime_per_task_per_bucket.div(evals_per_task) / 200

    required_runtime = model_ids.join(runtime_per_evaluation_per_epoch.rename("runtime"), on=fidelity_params)["runtime"]
    required_cpus = model_ids.join(cpus_per_task_per_bucket.rename("cpus"), on=fidelity_params)["cpus"]
    profile = pd.concat(
        {
            "required": pd.concat([required_runtime, required_cpus], axis=1),
            "model_config": model_ids,
            "job_config": basedirs
        },
        axis=1
    )

    return profile


if __name__ == "__main__":
    # test script
    # TODO: Adjust for new sampling-profile based operation
    # import sys
    #
    # fmt = logging.Formatter("[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S")
    # ch = logging.StreamHandler(stream=sys.stdout)
    # ch.setLevel(logging.DEBUG)
    # ch.setFormatter(fmt)
    # _log.addHandler(ch)
    # _log.setLevel(logging.DEBUG)
    #
    # modelid_level_names = [constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value]
    # modelids = pd.MultiIndex.from_product([[0, 1, 2], [0, 1, 2]], names=modelid_level_names)
    # runtimes = [100] * 9
    # nepochs = [300, 200, 200, 100, 100, 50, 50, 25, 25]
    # estimates = {}
    # max_epochs = 200
    #
    # subdfs = {}
    # for id, e, t in zip(modelids.values, nepochs, runtimes):
    #     tdata = [i * t / (e - 1) for i in range(e)]
    #     edata = [0.1 * t / e] * e
    #     subdfs[id] = pd.DataFrame({("diagnostic", "runtime"): tdata, ("train", "duration"): edata,
    #                                ("valid", "duration"): edata, ("test", "duration"): edata})
    #     estimates[id] = t / e * (max_epochs - e)
    #
    # metdf = pd.concat(subdfs, axis=0)
    # metdf.index = metdf.index.rename([*modelid_level_names, constants.MetricDFIndexLevels.epoch.value])
    # runtime_estimates = estimate_remaining_runtime(df=metdf, max_epochs=max_epochs)
    # assert runtime_estimates.columns.size == 1 and "required" in runtime_estimates.columns, \
    #     f"Unexpected runtime estimate DataFrame structure, with columns: {runtime_estimates.columns}"
    # assert runtime_estimates.index.size == len(modelids) and runtime_estimates.index.difference(modelids).size == 0, \
    #     f"Unexpected runtime estimate DataFrame structure.\nExpected model IDs: {modelids}\n" \
    #     f"DataFrame index:{runtime_estimates.index}"
    #
    # expected_estimates = pd.DataFrame(data=estimates, index=["required"]).transpose()
    # assert runtime_estimates.equals(expected_estimates), \
    #     f"Mismatch in expected runtime estimates and calculated estimates.\nExpected: {expected_estimates}\n" \
    #     f"Calculated: {runtime_estimates}"
    #
    # max_runtime = 700.
    # real_fidelity_params = fidelity_params
    # real_fidelity_types = fidelity_types
    # fidelity_params = ["C1", "C2"]
    # fidelity_types = {"C1": int, "C2": int}
    #
    # confs = {"C1": list(range(modelids.size)), "C2": list(range(modelids.size))[::-1]}
    # confs = pd.DataFrame(confs, index=modelids)
    # estimates_input_df = pd.concat({"model_config": confs, "runtime": runtime_estimates}, axis=1)
    # jobconf = JobConfig(cpus_per_worker=1, cpus_per_node=3, nodes_per_job=1, timelimit=300)
    # work_allocation = allocate_work(job_config=jobconf, runtime_estimates=estimates_input_df,
    #                                 cpuh_utilization_cutoff=0.75, dynamic_timelimit=False)
    # _log.info("This message should have been preceeded by a warning about low CPUh utilization, at about 0.33.")
    # assert len(work_allocation) == 9, f"Expected the work allocation to be split amongst 9 workers, was split " \
    #                                   f"amongst {len(work_allocation)} workers instead."
    # assert all([w.portfolio.shape[0] == 0 for w in work_allocation[3:]]), \
    #     f"Unexpected work allocation, work allocation for 6 of the 9 workers should have failed, was instead:\n" \
    #     f"{[str(w.portfolio) for w in work_allocation]}"
    #
    # jobconf = JobConfig(cpus_per_worker=1, cpus_per_node=3, nodes_per_job=1, timelimit=900)
    # work_allocation = allocate_work(job_config=jobconf, runtime_estimates=estimates_input_df,
    #                                 cpuh_utilization_cutoff=0.75, dynamic_timelimit=False)
    # _log.info("This message should not have been preceeded by a warning about low CPUh utilization.")
    # assert len(work_allocation) == 3, f"Expected the work allocation to be split amongst 3 workers, was split " \
    #                                   f"amongst {len(work_allocation)} workers instead."
    # assert jobconf.timelimit == 900, f"The timelimit of the job should not have been capped."
    #
    # jobconf = JobConfig(cpus_per_worker=1, cpus_per_node=3, nodes_per_job=1, timelimit=900)
    # work_allocation = allocate_work(job_config=jobconf, runtime_estimates=estimates_input_df,
    #                                 cpuh_utilization_cutoff=0.95, dynamic_timelimit=True)
    # _log.info("This message should have been preceeded by a warning about low CPUh utilization.")
    # assert len(work_allocation) == 3, f"Expected the work allocation to be split amongst 3 workers, was split " \
    #                                   f"amongst {len(work_allocation)} workers instead."
    # assert jobconf.timelimit == 800, f"The timelimit of the job should have been capped."
    #
    # fidelity_params = real_fidelity_params
    # fidelity_types = real_fidelity_types

    # Test resource estimation funcs

    fids = pd.MultiIndex.from_product([[1, ], [4, ], [0.25, 0.5, 1.0]], names=fidelity_params)
    cpus_per_bucket = pd.Series([1, 1, 4], index=fids)
    runtime_per_bucket = pd.Series([100, 400, 1600], index=fids)
    evals_per_worker = pd.Series([20, 20, 5], index=fids)
    test_nodes_per_bucket = pd.Series([5, 5, 20], index=fids)
    test_profile = prepare_full_sampling_profile(cpus_per_bucket, runtime_per_bucket, evals_per_worker,
                                                 test_nodes_per_bucket)

    assert len(test_profile.index) == 10800, f"Unexpected number of model configs: {len(test_profile.index)}"

    _log.info("Finished verification.")



""" Taken as-is under extant license conditions from the original NAS-Bench-201 repo's
code at
https://github.com/D-X-Y/AutoDL-Projects/tree/bc4c4692589e8ee7d6bab02603e69f8e5bd05edc """

import math, torch
import torch.nn as nn
from torch.optim import Optimizer

class _LRScheduler(object):

  def __init__(self, optimizer, warmup_epochs, epochs):
    if not isinstance(optimizer, Optimizer):
      raise TypeError('{:} is not an Optimizer'.format(type(optimizer).__name__))
    self.optimizer = optimizer
    for group in optimizer.param_groups:
      group.setdefault('initial_lr', group['lr'])
    self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))
    self.max_epochs = epochs
    self.warmup_epochs  = warmup_epochs
    self.current_epoch  = 0
    self.current_iter   = 0

  def extra_repr(self):
    return ''

  def __repr__(self):
    return ('{name}(warmup={warmup_epochs}, max-epoch={max_epochs}, current::epoch={current_epoch}, iter={current_iter:.2f}'.format(name=self.__class__.__name__, **self.__dict__)
              + ', {:})'.format(self.extra_repr()))

  def state_dict(self):
    return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

  def load_state_dict(self, state_dict):
    self.__dict__.update(state_dict)

  def get_lr(self):
    raise NotImplementedError

  def get_min_info(self):
    lrs = self.get_lr()
    return '#LR=[{:.6f}~{:.6f}] epoch={:03d}, iter={:4.2f}#'.format(min(lrs), max(lrs), self.current_epoch, self.current_iter)

  def get_min_lr(self):
    return min( self.get_lr() )

  def update(self, cur_epoch, cur_iter):
    if cur_epoch is not None:
      assert isinstance(cur_epoch, int) and cur_epoch>=0, 'invalid cur-epoch : {:}'.format(cur_epoch)
      self.current_epoch = cur_epoch
    if cur_iter is not None:
      assert isinstance(cur_iter, float) and cur_iter>=0, 'invalid cur-iter : {:}'.format(cur_iter)
      self.current_iter  = cur_iter
    for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
      param_group['lr'] = lr



class CosineAnnealingLR(_LRScheduler):

  def __init__(self, optimizer, warmup_epochs, epochs, T_max, eta_min):
    self.T_max = T_max
    self.eta_min = eta_min
    super(CosineAnnealingLR, self).__init__(optimizer, warmup_epochs, epochs)

  def extra_repr(self):
    return 'type={:}, T-max={:}, eta-min={:}'.format('cosine', self.T_max, self.eta_min)

  def get_lr(self):
    lrs = []
    for base_lr in self.base_lrs:
      if self.current_epoch >= self.warmup_epochs and self.current_epoch < self.max_epochs:
        last_epoch = self.current_epoch - self.warmup_epochs
        #if last_epoch < self.T_max:
        #if last_epoch < self.max_epochs:
        lr = self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * last_epoch / self.T_max)) / 2
        #else:
        #  lr = self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * (self.T_max-1.0) / self.T_max)) / 2
      elif self.current_epoch >= self.max_epochs:
        lr = self.eta_min
      else:
        lr = (self.current_epoch / self.warmup_epochs + self.current_iter / self.warmup_epochs) * base_lr
      lrs.append( lr )
    return lrs



import networkx as nx
import copy
import logging
import torch
import random
import inspect

from networkx.algorithms.dag import lexicographical_topological_sort

from jahs_bench.tabular.lib.naslib.utils.utils import iter_flatten, AttrDict
from jahs_bench.tabular.lib.naslib.utils.logging import log_formats, log_first_n
from .primitives import Identity, AbstractPrimitive
from .query_metrics import Metric

logger = logging.getLogger(__name__)


class Graph(torch.nn.Module, nx.DiGraph):
    """
    Base class for defining a search space. Add nodes and edges
    as for a directed acyclic graph in `networkx`. Nodes can contain
    graphs as children, also edges can contain graphs as operations.

    Note, if a graph is copied, the shared attributes of its edges are
    shallow copies whereas the private attributes are deep copies.

    To differentiate copies of the same graph you can define a `scope`
    with `set_scope()`.

    **Graph at nodes:**
    >>> graph = Graph()
    >>> graph.add_node(1, subgraph=Graph())

    If the node has more than one input use `set_input()` to define the
    routing to the input nodes of the subgraph.

    **Graph at edges:**
    >>> graph = Graph()
    >>> graph.add_nodes_from([1, 2])
    >>> graph.add_edge(1, 2, EdgeData({'op': Graph()}))

    **Modify the graph after definition**

    If you want to modify the graph e.g. in an optimizer once
    it has been defined already use the function `update_edges()`
    or `update_nodes()`.

    **Use as pytorch module**
    If you want to learn the weights of the operations or any
    other parameters of the graph you have to parse it first.
    >>> graph = getFancySearchSpace()
    >>> graph.parse()
    >>> logits = graph(data)
    >>> optimizer.min(loss(logits, target))

    To update the pytorch module representation (e.g. after removing or adding
    some new edges), you have to unparse. Beware that this is not fast, so it should
    not be done on each batch or epoch, rather once after discretizising. If you
    want to change the representation of the graph use rather some shared operation
    indexing at the edges.
    >>> graph.update(remove_random_edges)
    >>> graph.unparse()
    >>> graph.parse()
    >>> logits = graph(data)

    **Interface to tabular benchmarks**
    TODO
    """

    """
    Usually the optimizer does not operate on the whole graph, e.g. preprocessing
    and post-processing are excluded. Scope can be used to define that or to
    differentate instances of the "same" graph.
    """
    OPTIMIZER_SCOPE = "all"

    """
    Whether the search space has an interface to one of the tabular benchmarks which
    can then be used to query architecture performances.

    If this is set to true then `query()` should be implemented.
    """
    QUERYABLE = False

    def __init__(self, name: str = None, scope: str = None):
        """
        Initialise a graph. The edges are automatically filled with an EdgeData object
        which defines the default operation as Identity. The default combination operation
        is set as sum.

        Note:
            When inheriting form `Graph` note that `__init__()` cannot take any parameters.
            This is due to the way how networkx is implemented, i.e. graphs are reconstructed
            internally and no parameters for init are considered.

            Our recommended solution is to create static attributes before initialization and
            then load them dynamically in `__init__()`.

            >>> def __init__(self):
            >>>     num_classes = self.NUM_CLASSES
            >>> MyGraph.NUM_CLASSES = 42
            >>> my_graph_42_classes = MyGraph()

        """
        # super().__init__()
        nx.DiGraph.__init__(self)
        torch.nn.Module.__init__(self)

        # Make DiGraph a member and not inherit. This is because when inheriting from
        # `Graph` note that `__init__()` cannot take any parameters. This is due to
        # the way how networkx is implemented, i.e. graphs are reconstructed internally
        # and no parameters for init are considered.
        # Therefore __getattr__ and __iter__ forward the DiGraph methods for straight-forward
        # usage as if we would inherit.

        # self._nxgraph = nx.DiGraph()

        # Replace the default dicts at the edges with `EdgeData` objects
        # `EdgeData` can be easily customized and allow shared parameters
        # across different Graph instances.

        # self._nxgraph.edge_attr_dict_factory = lambda: EdgeData()
        self.edge_attr_dict_factory = lambda: EdgeData()

        # Replace the default dicts at the nodes to include `input` from the beginning.
        # `input` is required for storing the results of incoming edges.

        # self._nxgraph.node_attr_dict_factory = lambda: dict({'input': {}, 'comb_op': sum})
        self.node_attr_dict_factory = lambda: dict({"input": {}, "comb_op": sum})

        # remember to add all members also in `unparse()`
        self.name = name
        self.scope = scope
        self.input_node_idxs = None
        self.is_parsed = False
        self._id = random.random()  # pytorch expects unique modules in `add_module()`

    def __eq__(self, other):
        return self.name == other.name and self.scope == other.scope

    def __hash__(self):
        """
        As it is very complicated to compare graphs (i.e. check all edge
        attributes, do the have shared attributes, ...) use just the name
        for comparison.

        This is used when determining whether two instances are copies.
        """
        h = 0
        h += hash(self.name)
        h += hash(self.scope) if self.scope else 0
        h += hash(self._id)
        return h

    def __repr__(self):
        return "Graph {}-{:.07f}, scope {}, {} nodes".format(
            self.name, self._id, self.scope, self.number_of_nodes()
        )

    def modules_str(self):
        """
        Once the graph has been parsed, prints the modules as they appear in pytorch.
        """
        if self.is_parsed:
            result = ""
            for g in self._get_child_graphs(single_instances=True) + [self]:
                result += "Graph {}:\n {}\n==========\n".format(
                    g.name, torch.nn.Module.__repr__(g)
                )
            return result
        else:
            return self.__repr__()

    def set_scope(self, scope: str, recursively=True):
        """
        Sets the scope of this instance of the graph.

        The function should be used in a builder-like pattern
        `'subgraph'=Graph().set_scope("scope")`.

        Args:
            scope (str): the scope
            recursively (bool): Also set the scope for all child graphs.
                default True

        Returns:
            Graph: self with the setted scope.
        """
        self.scope = scope
        if recursively:
            for g in self._get_child_graphs(single_instances=False):
                g.scope = scope
        return self

    def add_node(self, node_index, **attr):
        """
        Adds a node to the graph.

        Note that adding a node using an index that has been used already
        will override its attributes.

        Args:
            node_index (int): The index for the node. Expect to be >= 1.
            **attr: The attributes which can be added in a dict like form.
        """
        assert node_index >= 1, "Expecting the node index to be greater or equal 1"
        nx.DiGraph.add_node(self, node_index, **attr)

    def copy(self):
        """
        Copy as defined in networkx, i.e. a shallow copy.

        Just handling recursively nested graphs seperately.
        """

        def copy_dict(d):
            copied_dict = d.copy()
            for k, v in d.items():
                if isinstance(v, Graph):
                    copied_dict[k] = v.copy()
                elif isinstance(v, list):
                    copied_dict[k] = [
                        i.copy() if isinstance(i, Graph) else i for i in v
                    ]
                elif isinstance(v, torch.nn.Module) or isinstance(v, AbstractPrimitive):
                    copied_dict[k] = copy.deepcopy(v)
            return copied_dict

        G = self.__class__()
        G.graph.update(self.graph)
        G.add_nodes_from((n, copy_dict(d)) for n, d in self._node.items())
        G.add_edges_from(
            (u, v, datadict.copy())
            for u, nbrs in self._adj.items()
            for v, datadict in nbrs.items()
        )
        G.scope = self.scope
        G.name = self.name
        return G

    def set_input(self, node_idxs: list):
        """
        Route the input from specific parent edges to the input nodes of
        this subgraph. Inputs are assigned in lexicographical order.

        Example:
        - Parent node (i.e. node where `self` is located on) has two
          incoming edges from nodes 3 and 5.
        - `self` has two input nodes 1 and 2 (i.e. nodes without
          an incoming edge)
        - `node_idxs = [5, 3]`
        Then input of node 5 is routed to node 1 and input of node 3
        is routed to node 2.

        Similarly, if `node_idxs = [5, 5]` then input of node 5 is routed
        to both node 1 and 2. Warning: In this case the output of another
        incoming edge is ignored!

        Should be used in a builder-like pattern: `'subgraph'=Graph().set_input([5, 3])`

        Args:
            node_idx (list): The index of the nodes where the data is coming from.

        Returns:
            Graph: self with input node indices set.

        """
        num_innodes = sum([self.in_degree(n) == 0 for n in self.nodes])
        assert num_innodes == len(
            node_idxs
        ), "Expecting node index for every input node. Excpected {}, got {}".format(
            num_innodes, len(node_idxs)
        )
        self.input_node_idxs = node_idxs
        return self

    def num_input_nodes(self) -> int:
        """
        The number of input nodes, i.e. the nodes without an
        incoming edge.

        Returns:
            int: Number of input nodes.
        """
        return sum(self.in_degree(n) == 0 for n in self.nodes)

    def _assign_x_to_nodes(self, x):
        """
        Assign x to the input nodes of self. Depending whether on
        edge or nodes.

        Performs also several sanity checks of the input.

        Args:
            x (Tensor or dict): Input to be assigned.
        """
        # We need dict in case of cell and int in case of motif
        assert isinstance(x, dict) or isinstance(x, torch.Tensor)

        if self.input_node_idxs is None:
            assert (
                self.num_input_nodes() == 1
            ), "There are more than one input nodes but input indeces are not defined."
            assert (
                len(list(self.predecessors(1))) == 0
            ), "Expecting node 1 to be the parent."
            assert (
                "subgraph" not in self.nodes[1].keys()
            ), "Expecting node 1 not to have a subgraph as it serves as input node."
            assert isinstance(x, torch.Tensor)
            self.nodes[1]["input"] = {0: x}
        else:
            # assign the input to the corresponding nodes
            assert all(
                [i in x.keys() for i in self.input_node_idxs]
            ), "got x from an unexpected input edge"
            if self.num_input_nodes() > len(x):
                # here is the case where the same input is assigned to more than one node
                # this can happen when there are cells with two inputs but at the very first
                # layer of the network, there is just one output (i.e. the data inputed to the
                # makro input node). Handle it and log a Info. This should happen only rarly
                logger.debug(
                    "We are using the same x for two inputs in graph {}".format(
                        self.name
                    )
                )
            input_node_iterator = iter(self.input_node_idxs)
            for node_idx in lexicographical_topological_sort(self):
                if self.in_degree(node_idx) == 0:
                    self.nodes[node_idx]["input"] = {0: x[next(input_node_iterator)]}

    def forward(self, x, *args):
        """
        Forward some data through the graph. This is done recursively
        in case there are graphs defined on nodes or as 'op' on edges.

        Args:
            x (Tensor or dict): The input. If the graph sits on a node the
                input can be a dict with {source_idx: Tensor} to be routed
                to the defined input nodes. If the graph sits on an edge,
                x is the feature tensor.
            args: This is only required to handle cases where the graph sits
                on an edge and receives an EdgeData object which will be ignored
        """
        logger.debug("Graph {} called. Input {}.".format(self.name, log_formats(x)))

        # Assign x to the corresponding input nodes
        self._assign_x_to_nodes(x)

        for node_idx in lexicographical_topological_sort(self):
            node = self.nodes[node_idx]
            logger.debug(
                "Node {}-{}, current data {}, start processing...".format(
                    self.name, node_idx, log_formats(node)
                )
            )

            # node internal: process input if necessary
            if ("subgraph" in node and "comb_op" not in node) or (
                "comb_op" in node and "subgraph" not in node
            ):
                log_first_n(
                    logging.WARN, "Comb_op is ignored if subgraph is defined!", n=1
                )
            # TODO: merge 'subgraph' and 'comb_op'. It is basicallly the same thing. Also in parse()
            if "subgraph" in node:
                x = node["subgraph"].forward(node["input"])
            else:
                if len(node["input"].values()) == 1:
                    x = list(node["input"].values())[0]
                else:
                    x = node["comb_op"](
                        [node["input"][k] for k in sorted(node["input"].keys())]
                    )
            node["input"] = {}  # clear the input as we have processed it

            if (
                len(list(self.neighbors(node_idx))) == 0
                and node_idx < list(lexicographical_topological_sort(self))[-1]
            ):
                # We have more than one output node. This is e.g. the case for
                # auxillary losses. Attach them to the graph, handling must done
                # by the user.
                logger.debug(
                    "Graph {} has more then one output node. Storing output of non-maximum index node {} at graph dict".format(
                        self, node_idx
                    )
                )
                self.graph["out_from_{}".format(node_idx)] = x
            else:
                # outgoing edges: process all outgoing edges
                for neigbor_idx in self.neighbors(node_idx):
                    edge_data = self.get_edge_data(node_idx, neigbor_idx)
                    # inject edge data only for AbstractPrimitive, not Graphs
                    if isinstance(edge_data.op, Graph):
                        edge_output = edge_data.op.forward(x)
                    elif isinstance(edge_data.op, AbstractPrimitive):
                        logger.debug(
                            "Processing op {} at edge {}-{}".format(
                                edge_data.op, node_idx, neigbor_idx
                            )
                        )

                        edge_output = edge_data.op.forward(x, edge_data=edge_data)
                    else:
                        raise ValueError(
                            "Unknown class as op: {}. Expected either Graph or AbstactPrimitive".format(
                                edge_data.op
                            )
                        )
                    self.nodes[neigbor_idx]["input"].update({node_idx: edge_output})

            logger.debug("Node {}-{}, processing done.".format(self.name, node_idx))

        logger.debug("Graph {} exiting. Output {}.".format(self.name, log_formats(x)))
        return x

    def parse(self):
        """
        Convert the graph into a neural network which can then
        be optimized by pytorch.
        """
        for node_idx in lexicographical_topological_sort(self):
            if "subgraph" in self.nodes[node_idx]:
                self.nodes[node_idx]["subgraph"].parse()
                self.add_module(
                    "{}-subgraph_at({})".format(self.name, node_idx),
                    self.nodes[node_idx]["subgraph"],
                )
            else:
                if isinstance(self.nodes[node_idx]["comb_op"], torch.nn.Module):
                    self.add_module(
                        "{}-comb_op_at({})".format(self.name, node_idx),
                        self.nodes[node_idx]["comb_op"],
                    )
            for neigbor_idx in self.neighbors(node_idx):
                edge_data = self.get_edge_data(node_idx, neigbor_idx)
                if isinstance(edge_data.op, Graph):
                    edge_data.op.parse()
                elif edge_data.op.get_embedded_ops():
                    for primitive in edge_data.op.get_embedded_ops():
                        if isinstance(primitive, Graph):
                            primitive.parse()
                self.add_module(
                    "{}-edge({},{})".format(self.name, node_idx, neigbor_idx),
                    edge_data.op,
                )
        self.is_parsed = True

    def unparse(self):
        """
        Undo the pytorch parsing by reconstructing the graph uusing the
        networkx data structures.

        This is done recursively also for child graphs.

        Returns:
            Graph: An unparsed shallow copy of the graph.
        """
        g = self.__class__()
        g.clear()

        graph_nodes = self.nodes
        graph_edges = self.edges

        # unparse possible child graphs
        # be careful with copying/deepcopying here cause of shared edge data
        for n, data in graph_nodes.data():
            if "subgraph" in data:
                data["subgraph"] = data["subgraph"].unparse()
        for u, v, data in graph_edges.data():
            if isinstance(data.op, Graph):
                data.set("op", data.op.unparse())

        # create the new graph
        # Remember to add all members here to update. I know it is ugly but don't know better
        g.add_nodes_from(graph_nodes.data())
        g.add_edges_from(graph_edges.data())
        g.graph.update(self.graph)
        g.name = self.name
        g.input_node_idxs = self.input_node_idxs
        g.scope = self.scope
        g.is_parsed = False
        g._id = self._id
        g.OPTIMIZER_SCOPE = self.OPTIMIZER_SCOPE
        g.QUERYABLE = self.QUERYABLE

        return g

    def _get_child_graphs(self, single_instances: bool = False) -> list:
        """
        Get all child graphs of the current graph.

        Args:
            single_instances (bool): Whether to return multiple instances
                (i.e. copies) of the same graph. When changing shared data
                this should be set to True.

        Returns:
            list: A list of all child graphs (can be empty)
        """
        graphs = []
        for node_idx in lexicographical_topological_sort(self):
            node_data = self.nodes[node_idx]
            if "subgraph" in node_data:
                graphs.append(node_data["subgraph"])
                graphs.append(node_data["subgraph"]._get_child_graphs())

        for _, _, edge_data in self.edges.data():
            if isinstance(edge_data.op, Graph):
                graphs.append(edge_data.op)
                graphs.append(edge_data.op._get_child_graphs())
            elif isinstance(edge_data.op, list):
                for op in edge_data.op:
                    if isinstance(op, Graph):
                        graphs.append(op)
                        graphs.append(op._get_child_graphs())
            elif isinstance(edge_data.op, AbstractPrimitive):
                # maybe it is an embedded op?
                embedded_ops = edge_data.op.get_embedded_ops()
                if embedded_ops is not None:
                    if isinstance(embedded_ops, Graph):
                        graphs.append(embedded_ops)
                        graphs.append(embedded_ops._get_child_graphs())
                    elif isinstance(embedded_ops, list):
                        for child_op in edge_data.op.get_embedded_ops():
                            if isinstance(child_op, Graph):
                                graphs.append(child_op)
                                graphs.append(child_op._get_child_graphs())
                    else:
                        logger.debug(
                            "Got embedded op, but is neither a graph nor a list: {}".format(
                                embedded_ops
                            )
                        )
            elif inspect.isclass(edge_data.op):
                assert not issubclass(
                    edge_data.op, Graph
                ), "Found non-initialized graph. Abort."
                pass  # we look at an uncomiled op
            else:
                raise ValueError("Unknown format of op: {}".format(edge_data.op))

        graphs = [g for g in iter_flatten(graphs)]

        if single_instances:
            single = []
            for g in graphs:
                if g.name not in [sg.name for sg in single]:
                    single.append(g)
            return sorted(single, key=lambda g: g.name)
        else:
            return sorted(graphs, key=lambda g: g.name)

    def get_all_edge_data(
        self, key: str, scope="all", private_edge_data: bool = False
    ) -> list:
        """
        Get edge attributes of this graph and all child graphs in one go.

        Args:
            key (str): The key of the attribute
            scope (str): The scope to be applied
            private_edge_data (bool): Whether to return data from graph copies as well.

        Returns:
            list: All data in a list.
        """
        assert scope is not None
        result = []
        for graph in self._get_child_graphs(single_instances=not private_edge_data) + [
            self
        ]:
            if (
                scope == "all"
                or graph.scope == scope
                or (isinstance(scope, list) and graph.scope in scope)
            ):
                for u, v, edge_data in graph.edges.data():
                    if edge_data.has(key):
                        result.append(edge_data[key])
        return result

    def set_at_edges(self, key, value, shared=False):
        """
        Sets the attribute for all edges in this and any child graph
        """
        for graph in self._get_child_graphs(single_instances=shared) + [self]:
            logger.debug("Updating edges of graph {}".format(graph.name))
            for _, _, edge_data in graph.edges.data():
                if not edge_data.is_final():
                    edge_data.set(key, value, shared)

    def compile(self):
        """
        Instanciates the ops at the edges using the arguments specified at the edges
        """
        for graph in self._get_child_graphs(single_instances=False) + [self]:
            logger.debug("Compiling graph {}".format(graph.name))
            for u, v, edge_data in graph.edges.data():
                if not edge_data.is_final():
                    attr = edge_data.to_dict()
                    op = attr.pop("op")

                    if isinstance(op, list):
                        compiled_ops = []
                        for i, o in enumerate(op):
                            if inspect.isclass(o):
                                # get the relevant parameter if there are more.
                                a = {
                                    k: v[i] if isinstance(v, list) else v
                                    for k, v in attr.items()
                                }
                                compiled_ops.append(o(**a))
                            else:
                                logger.debug(
                                    "op {} already compiled. Skipping".format(o)
                                )
                        edge_data.set("op", compiled_ops)
                    elif isinstance(op, AbstractPrimitive):
                        logger.debug("op {} already compiled. Skipping".format(op))
                    elif inspect.isclass(op) and issubclass(op, AbstractPrimitive):
                        # Init the class
                        edge_data.set("op", op(**attr))
                    elif isinstance(op, Graph):
                        pass  # This is already covered by _get_child_graphs
                    else:
                        raise ValueError("Unkown format of op: {}".format(op))

    def _verify_update_function(update_func: callable, private_edge_data: bool):
        """
        Verify that the update function actually modifies only
        shared/private edge data attributes based on setting of
        `private_edge_data`.

        Args:
            update_func (callable): callable that expects one argument
                named `current_edge_data`.
            private_edge_data (bool): Whether the update function is applied
                to all graph instances including copies or just to one instance
                per graph
        """

        test = EdgeData()
        test.set("shared", True, shared=True)
        test.set("op", [True])

        try:
            result = test.clone()
            update_func(current_edge_data=result)
        except:
            log_first_n(
                logging.WARN,
                "Update function could not be veryfied. Be cautious with the "
                "setting of `private_edge_data` in `update_edges()`",
                n=5,
            )
            return

        assert isinstance(
            result, EdgeData
        ), "Update function does not return the edge data object."

        if private_edge_data:
            assert result._shared == test._shared, (
                "The update function changes shared data although `private_edge_data` set to True. "
                "This is not the indended use of `update_edges`. The update function should only modify "
                "private edge data."
            )
        else:
            assert result._private == test._private, (
                "The update function changes private data although `private_edge_data` set to False. "
                "This is not the indended use of `update_edges`. The update function should only modify "
                "shared edge data."
            )

    def update_edges(
        self, update_func: callable, scope="all", private_edge_data: bool = False
    ):
        """
        This updates the edge data of this graph and all child graphs.
        This is the preferred way to manipulate the edges after the definition
        of the graph, e.g. by optimizers who want to insert their own op.
        `update_func(current_edge_data)`. This way optimizers
        can initialize and store necessary information at edges.

        Note that edges marked as 'final' will not be updated here.

        Args:
            update_func (callable): Function which accepts one argument called `current_edge_data`.
                and returns the modified EdgeData object.
            scope (str or list(str)): Can be "all" or list of scopes to be updated.
            private_edge_data (bool): If set to true, this means update_func will be
                applied to all edges. THIS IS NOT RECOMMENDED FOR SHARED
                ATTRIBUTES. Shared attributes should be set only once, we
                take care it is syncronized across all copies of this graph.

                The only usecase for setting it to true is when actually changing
                `op` during the initialization of the optimizer (e.g. replacing it
                with MixedOp or SampleOp)
        """
        Graph._verify_update_function(update_func, private_edge_data)
        assert scope is not None
        for graph in self._get_child_graphs(single_instances=not private_edge_data) + [
            self
        ]:
            if (
                scope == "all"
                or scope == graph.scope
                or (isinstance(scope, list) and graph.scope in scope)
            ):
                logger.debug("Updating edges of graph {}".format(graph.name))
                for u, v, edge_data in graph.edges.data():
                    if not edge_data.is_final():
                        edge = AttrDict(head=u, tail=v, data=edge_data)
                        update_func(edge=edge)
        self._delete_flagged_edges()

    def update_nodes(
        self, update_func: callable, scope="all", single_instances: bool = True
    ):
        """
        Update the nodes of the graph and its incoming and outgoing edges by iterating over the
        graph and applying `update_func` to each of it. This is the
        preferred way to change the search space once it has been defined.

        Note that edges marked as 'final' will not be updated here.

        Args:
            update_func (callable): Function that accepts three incoming parameters named
                `node, in_edges, out_edges`.
                    - `node` is a tuple (int, dict) containing the
                      index and the attributes of the current node.
                    - `in_edges` is a list of tuples with the index of
                      the tail of the edge and its EdgeData.
                    - `out_edges is a list of tuples with the index of
                      the head of the edge and its EdgeData.
            scope (str or list(str)): Can be "all" or list of scopes to be updated. Only graphs
                and child graphs with the specified scope are considered
            single_instance (bool): If set to false, this means update_func will be
                applied to nodes of all copies of a graphs. THIS IS NOT RECOMMENDED FOR SHARED
                ATTRIBUTES, i.e. when manipulating the shared data of incoming or outgoing edges.
                Shared attributes should be set only once, we take care it is syncronized across
                all copies of this graph.

                The only usecase for setting it to true is when actually changing
                `op` during the initialization of the optimizer (e.g. replacing it
                with MixedOp or SampleOp)
        """
        assert scope is not None
        for graph in self._get_child_graphs(single_instances) + [self]:
            if (
                scope == "all"
                or graph.scope == scope
                or (isinstance(scope, list) and graph.scope in scope)
            ):
                logger.debug("Updating nodes of graph {}".format(graph.name))
                for node_idx in lexicographical_topological_sort(graph):
                    node = (node_idx, graph.nodes[node_idx])
                    in_edges = list(graph.in_edges(node_idx, data=True))  # (v, u, data)
                    in_edges = [
                        (v, data) for v, u, data in in_edges if not data.is_final()
                    ]  # u is same for all
                    out_edges = list(
                        graph.out_edges(node_idx, data=True)
                    )  # (v, u, data)
                    out_edges = [
                        (u, data) for v, u, data in out_edges if not data.is_final()
                    ]  # v is same for all
                    update_func(node=node, in_edges=in_edges, out_edges=out_edges)
        self._delete_flagged_edges()

    def _delete_flagged_edges(self):
        """
        Delete edges which associated EdgeData is flagged as deleted.
        """
        for graph in self._get_child_graphs(single_instances=False) + [
            self
        ]:  # we operate on shallow copies
            to_remove = []
            for u, v, edge_data in graph.edges.data():
                if edge_data.is_deleted():
                    to_remove.append((u, v))
            if to_remove:
                # logger.info("Removing edges {} from graph {}".format(to_remove, graph))
                graph.remove_edges_from(to_remove)

    def clone(self):
        """
        Deep copy of the current graph.

        Returns:
            Graph: Deep copy of the graph.
        """
        return copy.deepcopy(self)

    def reset_weights(self, inplace: bool = False):
        """
        Resets the weights for the 'op' at all edges.

        Args:
            inplace (bool): Do the operation in place or
                return a modified copy.
        Returns:
            Graph: Returns the modified version of the graph.
        """

        def weight_reset(m):
            reset_parameters = getattr(m, "reset_parameters", None)
            if callable(reset_parameters):
                m.reset_parameters()

        if inplace:
            graph = self
        else:
            graph = self.clone()

        graph.apply(weight_reset)

        return graph

    def prepare_discretization(self):
        """
        In some cases the search space is manipulated before the final
        discretization is happening, e.g. DARTS. In such chases this should
        be defined in the search space, so all optimizers can call it.
        """
        pass

    def prepare_evaluation(self):
        """
        In some cases the evaluation architecture does not match the searched
        one. An example is where the makro_model is extended to increase the
        parameters. This is done here.
        """
        pass

    def sample_random_architecture(self, dataset_api):
        raise NotImplementedError()

    def mutate(self):
        raise NotImplementedError()

    def set_load_labeled(self):
        self.load_labeled = True

    def query(self, metric: Metric, dataset: str, path: str) -> float:
        """
        Can be used to query the performance of the architecture using
        a tabular benchmark.

        The interface must be provided by the search space.

        Args:
            metric (str): the name of the metric to query.
            dataset (str): the name of the dataset to query.
            path (str): path where the tabular data is located

        Returns:
            float: The performance number
        """
        if self.QUERYABLE:
            raise NotImplementedError(
                "QUERYABLE set to True therefore query_performance must be implemented"
            )
        else:
            raise NotImplementedError(
                "This function should not be used if QUERYABLE is False"
            )

    def get_dense_edges(self):
        """
        Returns the edge indices (i, j) that would make a fully connected
        DAG without circles such that i < j and i != j. Assumes nodes are
        already created.

        Returns:
            list: list of edge indices.
        """
        edges = []
        nodes = sorted(list(self.nodes()))
        for i in nodes:
            for j in nodes:
                if i != j and j > i:
                    edges.append((i, j))
        return edges

    def add_edges_densly(self):
        """
        Adds edges to get a fully connected DAG without cycles
        """
        self.add_edges_from(self.get_dense_edges())


class EdgeData:
    """
    Class that holds data for each edge.
    Data can be shared between instances of the graph
    where the edges lives in.

    Also defines the default key 'op', which is `Identity()`. It must
    be private always.

    Items can be accessed directly as attributes with `.key` or
    in a dict-like fashion with `[key]`. To set a new item use `.set()`.
    """

    def __init__(self, data={}):
        """
        Initializes a new EdgeData object.
        'op' is set as Identity() and private by default

        Args:
            data (dict): Inject some initial data. Will be always private.
        """
        self._private = {}
        self._shared = {}

        # set internal attributes
        self._shared["_deleted"] = False
        self._private["_final"] = False

        # set defaults and potential input
        self.set("op", Identity(), shared=False)
        for k, v in data.items():
            self.set(k, v, shared=False)

    def has(self, key: str):
        """
        Checks whether `key` exists.

        Args:
            key (str): The key to check.

        Returns:
            bool: True if key exists, False otherwise.

        """
        assert not key.startswith("_"), "Access to private keys not allowed!"
        if key in self._private.keys() or key in self._shared.keys():
            return True
        else:
            return False

    def __getitem__(self, key: str):
        assert not key.startswith("_"), "Access to private keys not allowed!"
        return self.__getattr__(key)

    def get(self, key: str, default):
        try:
            return self.__getattr__(key)
        except AttributeError:
            return default

    def __getattr__(self, key: str):
        if key.startswith("__"):  # Required for deepcopy, not sure why
            raise AttributeError(key)  #
        assert not key.startswith("_"), "Access to private keys not allowed!"
        if key in self._private:
            return self._private[key]
        elif key in self._shared:
            return self._shared[key]
        else:
            raise AttributeError(
                "Cannot find field '{}' in the given EdgeData!".format(key)
            )

    def __setattr__(self, name: str, val):
        if name.startswith("_"):
            super().__setattr__(name, val)
        else:
            raise ValueError("not allowed. use set().")

    def __str__(self):
        return "private: <{}>, shared: <{}>".format(
            str(self._private), str(self._shared)
        )

    def __repr__(self):
        return self.__str__()

    def update(self, data):
        """
        Update the data in here. If the data is added as dict,
        then all variables will be handled as private.

        Args:
            data (EdgeData or dict): If dict, then values will be set as
                private. If EdgeData then all entries will be replaced.
        """
        if isinstance(data, dict):
            for k, v in data.items():
                self.set(k, v)
        elif isinstance(data, EdgeData):
            # TODO: do update and not replace!
            self.__dict__.update(data.__dict__)
        else:
            raise ValueError("Unsupported type {}".format(data))

    def remove(self, key: str):
        """
        Removes an item from the EdgeData

        Args:
            key (str): The key for the item to be removed.
        """
        if key in self._private:
            del self._private[key]
        elif key in self._shared:
            del self._shared[key]
        else:
            raise KeyError("Tried to delete unkown key {}".format(key))

    def copy(self):
        """
        When a graph is copied to get multiple instances (e.g. when
        reusing subgraphs at more than one location) then
        this function will be called for all edges.

        It will create a deep copy for the private entries but
        only a shallow copy for the shared entries. E.g. architectural
        weights should be shared, but parameters of a 3x3 convolution not.

        Therefore 'op' must be always private.

        Returns:
            EdgeData: A new EdgeData object with independent private
                items, but shallow shared items.
        """
        new_self = EdgeData()
        new_self._private = copy.deepcopy(self._private)
        new_self._shared = self._shared

        # we need to handle copy of graphs seperately
        for k, v in self._private.items():
            if isinstance(v, Graph):
                new_self._private[k] = v.copy()
            elif isinstance(v, list):
                new_self._private[k] = [
                    i.copy() if isinstance(i, Graph) else i for i in v
                ]

        return new_self

    def set(self, key: str, value, shared=False):
        """
        Used to assign a new item to the EdgeData object.

        Args:
            key (str): The key.
            value (object): The value to store
            shared (bool): Default: False. Whether the item should
                be a shallow copy between different instances of EdgeData
                (and consequently between different instances of Graph).
        """
        assert isinstance(key, str), "Accepting only string keys, got {}".format(
            type(key)
        )
        assert not key.startswith("_"), "Access to private keys not allowed!"
        assert not self.is_final(), "Trying to change finalized edge!"
        if shared:
            if key in self._private:
                raise ValueError("Key {} alredy defined as non-shared")
            else:
                self._shared[key] = value
        else:
            if key in self._shared:
                raise ValueError("Key {} alredy defined as shared".format(key))
            else:
                self._private[key] = value

    def clone(self):
        """
        Return a true deep copy of EdgeData. Even shared
        items are not shared anymore.

        Returns:
            EdgeData: New independent instance.
        """
        return copy.deepcopy(self)

    def delete(self):
        """
        Flag to delete the edge where this instance is attached to.
        """
        self._shared["_deleted"] = True

    def is_deleted(self):
        """
        Returns true if the edge is flagged to be deleted
        """
        return self._shared["_deleted"]

    def finalize(self):
        """
        Sets this edge as final. This means it cannot be changed
        anymore and will also not appear in the update functions
        of the graph.
        """
        self._private["_final"] = True
        return self

    def is_final(self):
        """
        Returns:
            bool: True if the edge was finalized, False else
        """
        return self._private["_final"]

    def to_dict(self, subset="all"):
        if subset == "shared":
            return {k: v for k, v in self._shared.items() if not k.startswith("_")}
        elif subset == "private":
            return {k: v for k, v in self._private.items() if not k.startswith("_")}
        elif subset == "all":
            d = self.to_dict("private")
            d.update(self.to_dict("shared"))
            return d
        else:
            raise ValueError("Unknown subset {}".format(subset))



""" This file contains a number of constants that are or can be re-used throughout the search space definition. """

import enum
import torch.nn as nn

# Names of cell ops
OP_NAMES = ['Identity', 'Zero', 'ConvBN3x3', 'ConvBN1x1', 'AvgPool1x1']

# 2-tuples, (in-node, out-node), that uniquely identify cell edges
EDGE_LIST = ((1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4))

# Mapping from the original NASBench-201 dataset's naming convention to NASLib's naming convention
nb201_to_ops = {
    'avg_pool_3x3': 'AvgPool1x1',
    'nor_conv_1x1': 'ConvBN1x1',
    'nor_conv_3x3': 'ConvBN3x3',
    'skip_connect': 'Identity',
    'none': 'Zero',
}

# Mapping from NASLib's naming conversion to the original NASBench-201 dataset's naming convention
ops_to_nb201 = {
    'AvgPool1x1': 'avg_pool_3x3',
    'ConvBN1x1': 'nor_conv_1x1',
    'ConvBN3x3': 'nor_conv_3x3',
    'Identity': 'skip_connect',
    'Zero': 'none',
}

@enum.unique
class Activations(enum.Enum):
    ReLU = enum.auto(), nn.ReLU
    # SiLU = enum.auto(), nn.SiLU
    Hardswish = enum.auto(), nn.Hardswish
    Mish = enum.auto(), nn.Mish,
    # LeakyReLU = enum.auto(), nn.LeakyReLU



"""
Official implementation of TrivialAugment from the AutoML Repo:
https://github.com/automl/trivialaugment/tree/0eec449fdd71d9754e5317226917c6a8dcfcfee3
"""


import numpy as np
import re
from PIL import ImageOps, ImageEnhance, ImageFilter, Image, ImageDraw
import random
from dataclasses import dataclass
from typing import Union


@dataclass
class MinMax:
    min: Union[float, int]
    max: Union[float, int]


@dataclass
class MinMaxVals:
    shear: MinMax = MinMax(.0, .3)
    translate: MinMax = MinMax(0, 10)  # different from uniaug: MinMax(0,14.4)
    rotate: MinMax = MinMax(0, 30)
    solarize: MinMax = MinMax(0, 256)
    posterize: MinMax = MinMax(0, 4)  # different from uniaug: MinMax(4,8)
    enhancer: MinMax = MinMax(.1, 1.9)
    cutout: MinMax = MinMax(.0, .2)




def float_parameter(level, maxval):
    """Helper function to scale `val` between 0 and maxval .

  Args:
    level: Level of the operation that will be between [0, `PARAMETER_MAX`].
    maxval: Maximum value that the operation can have. This will be scaled
      to level/PARAMETER_MAX.

  Returns:
    A float that results from scaling `maxval` according to `level`.
  """
    return float(level) * maxval / PARAMETER_MAX


def int_parameter(level, maxval):
    """Helper function to scale `val` between 0 and maxval .

  Args:
    level: Level of the operation that will be between [0, `PARAMETER_MAX`].
    maxval: Maximum value that the operation can have. This will be scaled
      to level/PARAMETER_MAX.

  Returns:
    An int that results from scaling `maxval` according to `level`.
  """
    return int(level * maxval / PARAMETER_MAX)


class TransformFunction(object):
    """Wraps the Transform function for pretty printing options."""

    def __init__(self, func, name):
        self.f = func
        self.name = name

    def __repr__(self):
        return '<' + self.name + '>'

    def __call__(self, pil_img):
        return self.f(pil_img)


class TransformT(object):
    """Each instance of this class represents a specific transform."""

    def __init__(self, name, xform_fn):
        self.name = name
        self.xform = xform_fn

    def __repr__(self):
        return '<' + self.name + '>'

    def pil_transformer(self, probability, level):
        def return_function(im):
            if random.random() < probability:
                im = self.xform(im, level)
            return im

        name = self.name + '({:.1f},{})'.format(probability, level)
        return TransformFunction(return_function, name)


################## Transform Functions ##################
identity = TransformT('identity', lambda pil_img, level: pil_img)
flip_lr = TransformT(
    'FlipLR',
    lambda pil_img, level: pil_img.transpose(Image.FLIP_LEFT_RIGHT))
flip_ud = TransformT(
    'FlipUD',
    lambda pil_img, level: pil_img.transpose(Image.FLIP_TOP_BOTTOM))
# pylint:disable=g-long-lambda
auto_contrast = TransformT(
    'AutoContrast',
    lambda pil_img, level: ImageOps.autocontrast(
        pil_img))
equalize = TransformT(
    'Equalize',
    lambda pil_img, level: ImageOps.equalize(
        pil_img))
invert = TransformT(
    'Invert',
    lambda pil_img, level: ImageOps.invert(
        pil_img))
# pylint:enable=g-long-lambda
blur = TransformT(
    'Blur', lambda pil_img, level: pil_img.filter(ImageFilter.BLUR))
smooth = TransformT(
    'Smooth',
    lambda pil_img, level: pil_img.filter(ImageFilter.SMOOTH))


def _rotate_impl(pil_img, level):
    """Rotates `pil_img` from -30 to 30 degrees depending on `level`."""
    degrees = int_parameter(level, min_max_vals.rotate.max)
    if random.random() > 0.5:
        degrees = -degrees
    return pil_img.rotate(degrees)


rotate = TransformT('Rotate', _rotate_impl)


def _posterize_impl(pil_img, level):
    """Applies PIL Posterize to `pil_img`."""
    level = int_parameter(level, min_max_vals.posterize.max - min_max_vals.posterize.min)
    return ImageOps.posterize(pil_img, min_max_vals.posterize.max - level)


posterize = TransformT('Posterize', _posterize_impl)


def _shear_x_impl(pil_img, level):
    """Applies PIL ShearX to `pil_img`.

  The ShearX operation shears the image along the horizontal axis with `level`
  magnitude.

  Args:
    pil_img: Image in PIL object.
    level: Strength of the operation specified as an Integer from
      [0, `PARAMETER_MAX`].

  Returns:
    A PIL Image that has had ShearX applied to it.
  """
    level = float_parameter(level, min_max_vals.shear.max)
    if random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, level, 0, 0, 1, 0))


shear_x = TransformT('ShearX', _shear_x_impl)


def _shear_y_impl(pil_img, level):
    """Applies PIL ShearY to `pil_img`.

  The ShearY operation shears the image along the vertical axis with `level`
  magnitude.

  Args:
    pil_img: Image in PIL object.
    level: Strength of the operation specified as an Integer from
      [0, `PARAMETER_MAX`].

  Returns:
    A PIL Image that has had ShearX applied to it.
  """
    level = float_parameter(level, min_max_vals.shear.max)
    if random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, 0, level, 1, 0))


shear_y = TransformT('ShearY', _shear_y_impl)


def _translate_x_impl(pil_img, level):
    """Applies PIL TranslateX to `pil_img`.

  Translate the image in the horizontal direction by `level`
  number of pixels.

  Args:
    pil_img: Image in PIL object.
    level: Strength of the operation specified as an Integer from
      [0, `PARAMETER_MAX`].

  Returns:
    A PIL Image that has had TranslateX applied to it.
  """
    level = int_parameter(level, min_max_vals.translate.max)
    if random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, level, 0, 1, 0))


translate_x = TransformT('TranslateX', _translate_x_impl)


def _translate_y_impl(pil_img, level):
    """Applies PIL TranslateY to `pil_img`.

  Translate the image in the vertical direction by `level`
  number of pixels.

  Args:
    pil_img: Image in PIL object.
    level: Strength of the operation specified as an Integer from
      [0, `PARAMETER_MAX`].

  Returns:
    A PIL Image that has had TranslateY applied to it.
  """
    level = int_parameter(level, min_max_vals.translate.max)
    if random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, 0, 0, 1, level))


translate_y = TransformT('TranslateY', _translate_y_impl)


def _crop_impl(pil_img, level, interpolation=Image.BILINEAR):
    """Applies a crop to `pil_img` with the size depending on the `level`."""
    level = int_parameter(level, 10)
    w = pil_img.width
    h = pil_img.height
    cropped = pil_img.crop((level, level, w - level, h - level))
    resized = cropped.resize((w, h), interpolation)
    return resized


crop_bilinear = TransformT('CropBilinear', _crop_impl)


def _solarize_impl(pil_img, level):
    """Applies PIL Solarize to `pil_img`.

  Translate the image in the vertical direction by `level`
  number of pixels.

  Args:
    pil_img: Image in PIL object.
    level: Strength of the operation specified as an Integer from
      [0, `PARAMETER_MAX`].

  Returns:
    A PIL Image that has had Solarize applied to it.
  """
    level = int_parameter(level, min_max_vals.solarize.max)
    return ImageOps.solarize(pil_img, 256 - level)


solarize = TransformT('Solarize', _solarize_impl)


def _enhancer_impl(enhancer, minimum=None, maximum=None):
    """Sets level to be between 0.1 and 1.8 for ImageEnhance transforms of PIL."""

    def impl(pil_img, level):
        mini = min_max_vals.enhancer.min if minimum is None else minimum
        maxi = min_max_vals.enhancer.max if maximum is None else maximum
        v = float_parameter(level, maxi - mini) + mini  # going to 0 just destroys it
        return enhancer(pil_img).enhance(v)

    return impl


color = TransformT('Color', _enhancer_impl(ImageEnhance.Color))
ohl_color = TransformT('Color', _enhancer_impl(ImageEnhance.Color, .3, .9))
contrast = TransformT('Contrast', _enhancer_impl(ImageEnhance.Contrast))
brightness = TransformT('Brightness', _enhancer_impl(
    ImageEnhance.Brightness))
sharpness = TransformT('Sharpness', _enhancer_impl(ImageEnhance.Sharpness))
contour = TransformT(
    'Contour', lambda pil_img, level: pil_img.filter(ImageFilter.CONTOUR))
detail = TransformT(
    'Detail', lambda pil_img, level: pil_img.filter(ImageFilter.DETAIL))
edge_enhance = TransformT(
    'EdgeEnhance', lambda pil_img, level: pil_img.filter(ImageFilter.EDGE_ENHANCE))
sharpen = TransformT(
    'Sharpen', (lambda pil_img, level: pil_img.filter(ImageFilter.SHARPEN)))
max_ = TransformT(
    'Max', lambda pil_img, level: pil_img.filter(ImageFilter.MaxFilter))
min_ = TransformT(
    'Min', lambda pil_img, level: pil_img.filter(ImageFilter.MinFilter))
median = TransformT(
    'Median', lambda pil_img, level: pil_img.filter(ImageFilter.MedianFilter))
gaussian = TransformT(
    'Gaussian', lambda pil_img, level: pil_img.filter(ImageFilter.GaussianBlur))



def _mirrored_enhancer_impl(enhancer, minimum=None, maximum=None):
    """Sets level to be between 0.1 and 1.8 for ImageEnhance transforms of PIL."""

    def impl(pil_img, level):
        mini = min_max_vals.enhancer.min if minimum is None else minimum
        maxi = min_max_vals.enhancer.max if maximum is None else maximum
        assert mini == 0., "This enhancer is used with a strength space that is mirrored around one."
        v = float_parameter(level, maxi - mini) + mini  # going to 0 just destroys it
        if random.random() < .5:
            v = -v
        return enhancer(pil_img).enhance(1. + v)

    return impl


mirrored_color = TransformT('Color', _mirrored_enhancer_impl(ImageEnhance.Color))
mirrored_contrast = TransformT('Contrast', _mirrored_enhancer_impl(ImageEnhance.Contrast))
mirrored_brightness = TransformT('Brightness', _mirrored_enhancer_impl(
    ImageEnhance.Brightness))
mirrored_sharpness = TransformT('Sharpness', _mirrored_enhancer_impl(ImageEnhance.Sharpness))


def CutoutDefault(img, v):  # [0, 60] => percentage: [0, 0.2]
    # assert 0 <= v <= 20
    if v <= 0:
        return img
    w, h = img.size
    x0 = np.random.uniform(w)
    y0 = np.random.uniform(h)

    x0 = int(max(0, x0 - v / 2.))
    y0 = int(max(0, y0 - v / 2.))
    x1 = min(w, x0 + v)
    y1 = min(h, y0 + v)

    xy = (x0, y0, x1, y1)
    color = (0, 0, 0)
    img = img.copy()
    ImageDraw.Draw(img).rectangle(xy, color)
    return img


cutout = TransformT('Cutout',
                    lambda img, l: CutoutDefault(img, int_parameter(l, img.size[0] * min_max_vals.cutout.max)))




blend_images = None


def blend(img1, v):
    if blend_images is None:
        print("please set google_transformations.blend_images before using the enlarged_randaug search space.")
    i = np.random.choice(len(blend_images))
    img2 = blend_images[i]
    m = float_parameter(v, .4)
    return Image.blend(img1, img2, m)


sample_pairing = TransformT('SamplePairing', blend)


def set_augmentation_space(augmentation_space, num_strengths, custom_augmentation_space_augs=None):
    global ALL_TRANSFORMS, min_max_vals, PARAMETER_MAX
    assert num_strengths > 0
    PARAMETER_MAX = num_strengths - 1
    if 'wide' in augmentation_space:
        min_max_vals = MinMaxVals(
            shear=MinMax(.0, .99),
            translate=MinMax(0, 32),
            rotate=MinMax(0, 135),
            solarize=MinMax(0, 256),
            posterize=MinMax(2, 8),
            enhancer=MinMax(.01, 2.),
            cutout=MinMax(.0, .6),
        )
    elif ('uniaug' in augmentation_space) or ('randaug' in augmentation_space):
        min_max_vals = MinMaxVals(
            posterize=MinMax(4, 8),
            translate=MinMax(0, 14.4)
        )
    elif 'fixmirror' in augmentation_space:
        min_max_vals = MinMaxVals(
            posterize=MinMax(4, 8),
            enhancer=MinMax(0., .9)
        )
    elif 'fiximagenet' in augmentation_space:
        min_max_vals = MinMaxVals(
            posterize=MinMax(4, 8),
            translate=MinMax(0, 70)
        )
    elif 'fix' in augmentation_space:
        min_max_vals = MinMaxVals(
            posterize=MinMax(4, 8)
        )
    elif 'ohl' in augmentation_space:
        assert PARAMETER_MAX == 2
        min_max_vals = MinMaxVals(
            shear=MinMax(.1, .3),
            translate=MinMax(5, 14),
            rotate=MinMax(10, 30),
            solarize=MinMax(26, 179),
            posterize=MinMax(4, 7),
            enhancer=MinMax(1.3, 1.9),
            cutout=MinMax(.0, .6),
        )
    else:
        min_max_vals = MinMaxVals()

    if 'xlong' in augmentation_space:
        ALL_TRANSFORMS = [
            identity,
            auto_contrast,
            equalize,
            rotate,
            solarize,
            color,
            posterize,
            contrast,
            brightness,
            sharpness,
            shear_x,
            shear_y,
            translate_x,
            translate_y,
            blur,
            invert,
            flip_lr,
            flip_ud,
            cutout,
            crop_bilinear,
            contour,
            detail,
            edge_enhance,
            sharpen,
            max_,
            min_,
            median,
            gaussian
        ]
    elif 'rasubsetof' in augmentation_space:
        r = re.findall(r'rasubsetof(\d+)', augmentation_space)
        assert len(r) == 1
        ALL_TRANSFORMS = random.sample(ALL_TRANSFORMS, int(r[0]))
        print(f"Subsampled {len(ALL_TRANSFORMS)} augs: {ALL_TRANSFORMS}")
    elif 'fixmirror' in augmentation_space:
        ALL_TRANSFORMS = [
            identity,
            auto_contrast,
            equalize,
            rotate,
            solarize,
            mirrored_color,  # enhancer
            posterize,
            mirrored_contrast,  # enhancer
            mirrored_brightness,  # enhancer
            mirrored_sharpness,  # enhancer
            shear_x,
            shear_y,
            translate_x,
            translate_y
        ]
    elif 'long' in augmentation_space:
        ALL_TRANSFORMS = [
            identity,
            auto_contrast,
            equalize,
            rotate,
            solarize,
            color,
            posterize,
            contrast,
            brightness,
            sharpness,
            shear_x,
            shear_y,
            translate_x,
            translate_y,
            # sample_pairing,
            blur,
            invert,
            flip_lr,
            flip_ud,
            cutout
        ]
    elif 'uniaug' in augmentation_space:
        ALL_TRANSFORMS = [
            identity,
            shear_x,
            shear_y,
            translate_x,
            translate_y,
            rotate,
            auto_contrast,
            invert,  # only uniaug
            equalize,
            solarize,
            posterize,
            contrast,
            color,
            brightness,
            sharpness,
            cutout  # only uniaug
        ]
    elif 'autoaug_paper' in augmentation_space:
        ALL_TRANSFORMS = [
            shear_x,
            shear_y,
            translate_x,
            translate_y,
            rotate,
            auto_contrast,
            invert,
            equalize,
            solarize,
            posterize,
            contrast,
            color,
            brightness,
            sharpness,
            cutout,
            sample_pairing
        ]
    elif 'full' in augmentation_space:
        ALL_TRANSFORMS = [
            flip_lr,
            flip_ud,
            auto_contrast,
            equalize,
            invert,
            rotate,
            posterize,
            crop_bilinear,
            solarize,
            color,
            contrast,
            brightness,
            sharpness,
            shear_x,
            shear_y,
            translate_x,
            translate_y,
            cutout,
            blur,
            smooth
        ]
    elif 'ohl' in augmentation_space:
        ALL_TRANSFORMS = [
            shear_x,  # ok
            shear_y,  # ok
            translate_x,  # ok
            translate_y,  # ok
            rotate,  # ok
            ohl_color,  # nok
            posterize,  # ok
            solarize,  # ok
            contrast,  # ok
            sharpness,  # ok
            brightness,  # ok
            auto_contrast,
            equalize,
            invert
        ]
    elif 'custom' in augmentation_space:
        assert custom_augmentation_space_augs is not None

        custom_augmentation_space_augs_mapping = {
            'identity': identity,
            'auto_contrast': auto_contrast,
            'equalize': equalize,
            'rotate': rotate,
            'solarize': solarize,
            'color': color,
            'posterize': posterize,
            'contrast': contrast,
            'brightness': brightness,
            'sharpness': sharpness,
            'shear_x': shear_x,
            'shear_y': shear_y,
            'translate_x': translate_x,
            'translate_y': translate_y,
            # sample_pairing,
            'blur': blur,
            'invert': invert,
            'flip_lr': flip_lr,
            'flip_ud': flip_ud,
            'cutout': cutout,
            'crop_bilinear': crop_bilinear,
            'contour': contour,
            'detail': detail,
            'edge_enhance': edge_enhance,
            'sharpen': sharpen,
            'max_': max_,
            'min_': min_,
            'median': median,
            'gaussian': gaussian
        }
        ALL_TRANSFORMS = []
        ALL_TRANSFORMS += [
            custom_augmentation_space_augs_mapping[aug] for aug in custom_augmentation_space_augs
        ]
        print("CUSTOM Augs set to:", ALL_TRANSFORMS)
    else:
        if 'standard' not in augmentation_space:
            raise ValueError(f"Unknown search space {augmentation_space}")
        ALL_TRANSFORMS = [
            identity,
            auto_contrast,
            equalize,
            rotate,  # extra coin-flip
            solarize,
            color,  # enhancer
            posterize,
            contrast,  # enhancer
            brightness,  # enhancer
            sharpness,  # enhancer
            shear_x,  # extra coin-flip
            shear_y,  # extra coin-flip
            translate_x,  # extra coin-flip
            translate_y  # extra coin-flip
        ]

set_augmentation_space('fixed_standard', 31)


def apply_augmentation(aug_idx, m, img):
    return ALL_TRANSFORMS[aug_idx].pil_transformer(1., m)(img)


def num_augmentations():
    return len(ALL_TRANSFORMS)


class TrivialAugment:
    def __call__(self, img):
        op = random.choices(ALL_TRANSFORMS, k=1)[0]
        level = random.randint(0, PARAMETER_MAX)
        img = op.pil_transformer(1., level)(img)
        return img


class RandAugment:
    def __init__(self, n, m):
        self.n = n
        self.m = m  # [0, 30]

    def __call__(self, img):
        ops = random.choices(ALL_TRANSFORMS, k=self.n)
        for op in ops:
            img = op.pil_transformer(1., self.m)(img)

        return img


class UniAugment:
    def __call__(self, img):
        ops = random.choices(ALL_TRANSFORMS, k=2)
        for op in ops:
            level = random.randint(0, PARAMETER_MAX)
            img = op.pil_transformer(0.5, level)(img)
        return img


class UniAugmentWeighted:
    def __init__(self, n, probs):
        self.n = n
        self.probs = probs  # [prob of zero augs, prob of one aug, ..]

    def __call__(self, img):
        k = random.choices(range(len(self.probs)), self.probs)[0]
        ops = random.choices(ALL_TRANSFORMS, k=k)
        for op in ops:
            level = random.randint(0, PARAMETER_MAX)
            img = op.pil_transformer(1., level)(img)
        return img



from enum import Enum, auto


class Metric(Enum):
    RAW = auto()
    ALL = auto()

    TRAIN_ACCURACY = auto()
    VAL_ACCURACY = auto()
    TEST_ACCURACY = auto()

    TRAIN_LOSS = auto()
    VAL_LOSS = auto()
    TEST_LOSS = auto()

    TRAIN_TIME = auto()
    VAL_TIME = auto()
    TEST_TIME = auto()

    FLOPS = auto()
    LATENCY = auto()
    PARAMETERS = auto()
    EPOCH = auto()
    HP = auto()



import argparse
import logging
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
import pandas as pd
# Make sure that the system path contains the correct repository
from jahs_bench.surrogate import model, utils

_seed = 3501623856

_log = logging.getLogger(__name__)


def generate_splits(datadir: Path, datafile: str, test_size: float,
                    valid_size: Optional[float] = None,
                    outputs: Optional[Sequence[str]] = None, seed: Optional[int] = _seed):
    _log.info(f"Setting up the output directory and random state using the seed {seed}.")
    subdir = datadir / f"valid-{valid_size}-test-{test_size}"
    subdir.mkdir(exist_ok=False, parents=False)
    random_state = np.random.RandomState(seed)

    _log.info("Loading dataset from disk.")
    datapth = datadir / datafile
    features, labels, groups, strata = utils.adjust_dataset(
        datapth=datapth, outputs=outputs, fillna=True)

    if valid_size > 0.:
        train_set_frac = 1 - min(test_size, 1 - test_size)
        valid_frac = valid_size / train_set_frac
        _log.info(f"Using {valid_frac * 100} % of the training set to achieve an "
                  f"effective validation set fraction of {valid_size}.")
    else:
        valid_frac = 0.2  # Placeholder, this won't actually generate a valid split

    _log.info(f"Generating training and test splits using test set fraction {test_size}.")
    if test_size > 0.5:
        xtest, xtrain, ytest, ytrain, groups, strata, cv = \
            model.XGBSurrogate.prepare_dataset_for_training(
                features=features, labels=labels, groups=groups, test_size=1 - test_size,
                random_state=random_state, num_cv_splits=int(1 / valid_frac),
                stratify=True, strata=strata
            )
    else:
        xtrain, xtest, ytrain, ytest, groups, strata, cv = \
            model.XGBSurrogate.prepare_dataset_for_training(
                features=features, labels=labels, groups=groups, test_size=test_size,
                random_state=random_state, num_cv_splits=int(1 / valid_frac),
                stratify=True, strata=strata
            )

    _log.info(f"The test split contains {xtest.shape[0]} data points.")

    if valid_size > 0.:
        _log.info(f"Generating validation split.")
        idx_train, idx_valid = next(cv.split(xtrain, strata, groups=groups))
        xvalid = xtrain.iloc[idx_valid]
        yvalid = ytrain.iloc[idx_valid]
        xtrain = xtrain.iloc[idx_train]
        ytrain = ytrain.iloc[idx_train]
        groups = groups.iloc[idx_train]
        strata = strata.iloc[idx_train]
        _log.info(f"The training and validation splits contain {xtrain.shape[0]} and "
                  f"{xvalid.shape[0]} data points respectively.")
    else:
        _log.info(f"The training split contains {xtrain.shape[0]} data points.")

    _log.info("Saving training, validation and test sets to disk")
    pd.concat({
        "features": xtrain, "labels": ytrain,
        "groups": groups, "strata": strata
    }, axis=1).to_pickle(subdir / "train_set.pkl.gz")

    pd.concat({"features": xvalid, "labels": yvalid}, axis=1).to_pickle(
        subdir / "valid_set.pkl.gz")
    pd.concat({"features": xtest, "labels": ytest}, axis=1).to_pickle(
        subdir / "test_set.pkl.gz")

    _log.info(f"Finished generating splits for test size {test_size}.")


def parse_cli():
    parser = argparse.ArgumentParser(
        "Script to prepare the performance dataset for use in training a surrogate model."
    )
    parser.add_argument("--datadir", type=Path,
                        help="The directory where all the relevant data is stored as "
                             "well as where the data splits will be stored.")
    parser.add_argument("--datafile", type=str,
                        help="The name of the file inside `datadir` which contains the "
                             "full dataset that should be split up.")
    parser.add_argument("--test_size", type=float,
                        help="The fraction of the full dataset that should be used for "
                             "generating the test split.")
    parser.add_argument("--valid_size", type=float,
                        help="The fraction of the full dataset that should be used for "
                             "generating the validation split.")
    parser.add_argument("--outputs", type=str, default=None,
                        nargs=argparse.REMAINDER,
                        help="Strings, separated by spaces, that indicate which of the "
                             "available metrics in the dataset should be used for "
                             "generating the splits. If not given, all metrics present "
                             "in the full dataset are used.")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_cli()
    logging.basicConfig(level=logging.INFO,
                        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
                        datefmt="%m/%d %H:%M:%S")
    _log.setLevel(logging.INFO)
    generate_splits(**vars(args))




import argparse
import logging
import shutil
import sys
import time
from pathlib import Path
from typing import Generator, Tuple

import pandas as pd

from jahs_bench.tabular.lib.core import constants
from jahs_bench.tabular.lib.core import utils
from jahs_bench.tabular.lib.postprocessing import verification
from jahs_bench.tabular.clusterlib.prescheduler import fidelity_basedir_map

_log = logging.getLogger(__name__)
fidelity_parameters = ["N", "W", "Resolution"]
fidelity_to_str = {"N": lambda x: str(int(x)), "W": lambda x: str(int(x)), "Resolution": lambda x: str(x)}
worker_chkpt_subdir = "worker_chkpts"
cleanup_chkpt_name = "cleanup_progress.pkl.gz"
prune_chkpt_name = "prune_progress.pkl.gz"
verify_chkpt_name = "verify_progress.pkl.gz"

modes_of_operation = ["prepare", "cleanup", "prune", "verify"]


def generate_worker_portfolio(workerid: int, nworkers: int, configs_pth: Path) -> pd.DataFrame:
    configs: pd.DataFrame = pd.read_pickle(configs_pth)
    _log.debug(f"Found {configs.index.size} configurations to construct portfolio from.")

    assert isinstance(configs, pd.DataFrame), f"The configs path should lead to a DataFrame, was {type(configs)}."
    assert isinstance(configs.index, pd.MultiIndex), \
        f"The configs DataFrame's index has not been configured properly. It should be a MultiIndex, was " \
        f"{type(configs.index)}"
    assert constants.MetricDFIndexLevels.taskid.value in configs.index.names, \
        f"Configs DataFrame index has not been configured properly, could not find " \
        f"{constants.MetricDFIndexLevels.taskid.value}"
    assert constants.MetricDFIndexLevels.modelid.value in configs.index.names, \
        f"Configs DataFrame index has not been configured properly, could not find " \
        f"{constants.MetricDFIndexLevels.modelid.value}"

    portfolio = configs.loc[configs.index[workerid::nworkers], :]
    _log.debug(f"Created portfolio of {portfolio.index.size} configs.")
    return portfolio


def _iterate_portfolio(portfolio: pd.DataFrame) -> Generator[Tuple[Tuple[int, int], pd.Series], None, None]:
    """ Generator function that, given a portfolio in the form of a pandas DataFrame, iterates over the model
    identifiers and the model configuration. The identifiers are returned as (taskid, model_idx) and the configuration
    as a Pandas Series containing the various parameter names in the index. """

    taskid_pos = portfolio.index.names.index(constants.MetricDFIndexLevels.taskid.value)
    modelidx_pos = portfolio.index.names.index(constants.MetricDFIndexLevels.modelid.value)

    for i in range(portfolio.index.size):
        current_config = portfolio.index[i]
        model_id = int(current_config[taskid_pos]), int(current_config[modelidx_pos])
        yield model_id, portfolio.loc[current_config, :]


def subdir_from_fidelity(config: pd.Series):
    subdir = Path("-".join(["-".join([f, fidelity_to_str[f](config[f])]) for f in fidelity_parameters])) / "tasks"
    return subdir


def prepare_for_verification(basedir: Path, backupdir: Path):

    backupdir.mkdir(exist_ok=True, parents=True)
    try:
        backuptree = utils.DirectoryTree(basedir=backupdir)
    except FileExistsError:
        pass
    _log.debug(f"Preparing directory tree in {backuptree.basedir}.")
    for dtree in verification.iterate_model_tree(basedir=basedir, enumerate=False):
        try:
            backuptree.taskid = dtree.taskid
        except FileExistsError:
            pass
        try:
            backuptree.model_idx = dtree.model_idx
        except FileExistsError:
            pass

    _log.info("Finished creating backup directory structure.")



def clean_data(portfolio: pd.DataFrame, rootdir: Path, backupdir: Path, worker_chkpt_dir: Path, budget: int):
    _log.info(f"Performing clean up operation on data stored in root directory {rootdir}.")

    ## Prepare checkpoint
    chkpt_filename = worker_chkpt_dir / cleanup_chkpt_name
    done = None
    if chkpt_filename.exists():
        try:
            done = pd.read_pickle(chkpt_filename)
        except Exception as e:
            _log.info(f"Ran into an error while trying to read {chkpt_filename}: {str(e)}")
            done = None

    if done is None:
        done = pd.Series(False, index=portfolio.index).reorder_levels(
            [constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value])
        done.to_pickle(chkpt_filename)

    ## Estimated time budget requirement
    max_time = 0.
    start = time.time()
    for (taskid, model_idx), config in _iterate_portfolio(portfolio):
        if budget < 1.2 * max_time:
            break

        if (taskid, model_idx) in done.index and done[(taskid, model_idx)]:
            continue

        subdir = subdir_from_fidelity(config)
        _log.debug(f"Performing clean up for task {taskid}, model {model_idx} in directory tree at {subdir}.")
        verification.clean_corrupt_files(basedir=rootdir / subdir, taskid=taskid, model_idx=model_idx, cleanup=True,
                                         backupdir=backupdir / subdir)

        ## Safely update checkpoint
        done[(taskid, model_idx)] = True
        shutil.copyfile(chkpt_filename, f"{chkpt_filename}.bak")
        done.to_pickle(chkpt_filename)

        ## Update duration estimate
        end = time.time()
        duration = end - start
        max_time = max(max_time, end - start)
        budget -= duration
        start = end

    _log.info(f"Finished clean up operation, max operation time: {max_time} seconds.")


def prune_data(portfolio: pd.DataFrame, rootdir: Path, backupdir: Path, worker_chkpt_dir: Path, budget: int):
    _log.info(f"Performing data pruning operation on data stored in root directory {rootdir}.")

    ## Prepare checkpoint
    chkpt_filename = worker_chkpt_dir / prune_chkpt_name
    done = None
    if chkpt_filename.exists():
        try:
            done = pd.read_pickle(chkpt_filename)
        except Exception as e:
            _log.info(f"Ran into an error while trying to read {chkpt_filename}: {str(e)}")
            done = None

    if done is None:
        done = pd.Series(False, index=portfolio.index).reorder_levels(
            [constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value])
        done.to_pickle(chkpt_filename)

    ## Estimated time budget requirement
    max_time = 0.
    start = time.time()
    for (taskid, model_idx), config in _iterate_portfolio(portfolio):
        if budget < 1.2 * max_time:
            break

        if (taskid, model_idx) in done.index and done[(taskid, model_idx)]:
            continue

        subdir = subdir_from_fidelity(config)
        _log.debug(f"Pruning out-of-sync data for task {taskid}, model {model_idx} in directory tree at "
                   f"{rootdir / subdir}.")
        verification.check_metric_data_integrity(backup_dir=backupdir / subdir, basedir=rootdir / subdir,
                                                 taskid=taskid, model_idx=model_idx, cleanup=True)

        ## Safely update checkpoint
        done[(taskid, model_idx)] = True
        shutil.copyfile(chkpt_filename, f"{chkpt_filename}.bak")
        done.to_pickle(chkpt_filename)

        ## Update duration estimate
        end = time.time()
        duration = end - start
        max_time = max(max_time, end - start)
        budget -= duration
        start = end

    _log.info(f"Finished pruning operation, max operation time: {max_time} seconds.")


def verify_data_integrity(portfolio: pd.DataFrame, rootdir: Path, backupdir: Path, nepochs: int):
    pass


def argument_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("mode", choices=modes_of_operation,
                        help="Mode of operation for this script. There are four modes of operation: 'prepare', "
                             "'cleanup', 'prune' and 'verify'. In most use-cases, this script should be run in each of "
                             "these modes exactly in the order of their appearance above. 'prepare' runs a series of "
                             "preparatory procedures, such as creating the backup directory tree. 'cleanup' performs a "
                             "very simple and quick check to remove all checkpoints and metrics logs that are "
                             "unreadable. 'prune' is more involved than 'cleanup' and checks the checkpoints and "
                             "metrics logs for consistency. Any such pairs of files that are found to not be in sync "
                             "with each other, as well as subsequent pairs that would be affected, are removed. If "
                             "either of these two operations results in changes to the stored data, the respective "
                             "model's directory is marked as needing verification. In such a case, the 'verify' mode "
                             "of this script comes into play. It loads each checkpoint as well as the next metrics "
                             "log in chronologically ascending order, performs 'nepochs' evaluations from that "
                             "checkpoint onwards, and compares the generated data against the logged metrics data to "
                             "verify whether or not this checkpoint and metric log are valid. If an inconsistency is "
                             "found, all data files starting from the faulty checkpoint are removed.")
    parser.add_argument("--rootdir", type=Path, default=Path().cwd(),
                        help="Path to the root directory where all the tasks' output was stored. Task-specific "
                             "sub-directories will be created here if needed. Note that this is NOT the same as the "
                             "'basedir' of a DirectoryTree object as multiple DirectoryTree objects will be created "
                             "as and when needed.")
    parser.add_argument("--backupdir", type=Path, default="/tmp/nashpo_benchmarks/backupdir",
                        help="For the modes 'prune' and 'verify' of this script, data is never automatically deleted. "
                             "Removal only refers to moving the faulty data to an identical directory tree in this "
                             "backup directory. It must then be manually deleted when it has been verified that the "
                             "files are truly no longer necessary.")
    parser.add_argument("--workerid", type=int,
                        help="An offset from 0 for this worker within the current job's allocation of workers. This "
                             "does NOT correspond to a single 'taskid' but rather the portfolio of configurations that "
                             "this worker will evaluate.")
    parser.add_argument("--workerid-offset", type=int, default=0,
                        help="An additional fixed offset from 0 for this worker that, combined with the value of "
                             "'--worker', gives the overall worker ID in the context of a job that has been split into "
                             "multiple smaller parts.")
    parser.add_argument("--nworkers", type=int,
                        help="The total number of workers that are expected to concurrently perform data verification "
                             "in this root directory. This is used to coordinate the workers such that there is no "
                             "overlap in their portfolios.")
    parser.add_argument("--datadir", type=Path,
                        help="The directory where all datasets are expected to be stored.")
    parser.add_argument("--configs", type=Path,
                        help="Path to a pickled Pandas DataFrame (*.pkl.gz) that contains the full configuration of "
                             "each known model that should be verified. The index of each such configuration should be "
                             "the unique tuple (taskid, model_idx), where the corresponding values will be used to "
                             "construct corresponding directory trees.")
    parser.add_argument("--nepochs", type=int, default=None,
                        help="Only used in the 'verify' mode of operation. The number of epochs of data that should be "
                             "generated and verified using each loaded checkpoint. This implies that, for example, in "
                             "a model with 5 pairs of checkpoints and metric logs that have already been verified to "
                             "be in sync, the first 4 checkpoints will each be loaded up, the model trained for "
                             "'nepochs' epochs, and the corresponding entries in the last 4 metrics logs will be "
                             "cross-checked.")
    parser.add_argument("--budget", type=int,
                        help="Time budget for each worker in seconds. This is used to approximate whether or not a "
                             "worker has enough time to perform one more operation without any risk to any data.")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")

    return parser


if __name__ == "__main__":

    # Setup this module's logger
    fmt = logging.Formatter("[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S")
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(fmt)
    _log.addHandler(ch)
    _log.setLevel(logging.INFO)

    verification._log.addHandler(ch)
    verification._log.setLevel(logging.INFO)

    ## Parse CLI
    args = argument_parser().parse_args()

    if args.debug:
        _log.setLevel(logging.DEBUG)
        verification._log.setLevel(logging.DEBUG)

    if not args.rootdir.exists():
        raise RuntimeError(f"The given root directory {args.rootdir} does not exist.")

    workerid = args.workerid_offset + args.workerid
    portfolio = generate_worker_portfolio(workerid=workerid, nworkers=args.nworkers, configs_pth=args.configs)
    worker_chkpt_dir: Path = args.rootdir / worker_chkpt_subdir / str(workerid)
    worker_chkpt_dir.mkdir(exist_ok=True, parents=True)

    _log.info(f"Worker {workerid + 1}: Beginning.")

    if args.mode == modes_of_operation[0]:
        basedirs = portfolio.loc[:, fidelity_parameters].apply(fidelity_basedir_map, axis=1).rename("basedir")
        basedirs = basedirs.unique()
        if workerid + 1 > basedirs.size:
            print(f"Worker {workerid + 1}: No preparatory work to do given that there are {basedirs.size} possible "
                  f"fidelity groupy defined in the given profile.")
            sys.exit(0)
        else:
            basedir = args.rootdir / basedirs[workerid]
            backupdir = args.backupdir / basedirs[workerid]
            _log.info(f"Worker {workerid + 1}: Preparing backup directory structure at {backupdir} for {basedir}.")

        prepare_for_verification(basedir=basedir, backupdir=backupdir)
    elif args.mode == modes_of_operation[1]:
        clean_data(portfolio, rootdir=args.rootdir, backupdir=args.backupdir, worker_chkpt_dir=worker_chkpt_dir,
                   budget=args.budget)
    elif args.mode == modes_of_operation[2]:
        prune_data(portfolio, rootdir=args.rootdir, backupdir=args.backupdir, worker_chkpt_dir=worker_chkpt_dir,
                   budget=args.budget)
    else:
        raise NotImplementedError(f"Worker {workerid + 1}: No existing implementation for mode {args.mode}.")

    _log.info(f"Worker {workerid + 1}: Finished.")



import dataclasses as DC
from enum import Enum
from functools import partial
from typing import Union, Sequence

import pandas as pd
from scipy.stats import kendalltau
from sklearn.metrics import r2_score, mean_squared_error, \
    mean_absolute_percentage_error, median_absolute_error

import logging
_log = logging.getLogger(__name__)

# Some scaffolding for handling fidelity values semantically and fluidly
@DC.dataclass(init=True, eq=True, frozen=True, repr=True, order=False, unsafe_hash=True)
class Fidelity:
    Resolution: Union[Sequence[int], slice] = DC.field(default_factory=list)
    W: Union[Sequence[int], slice] = DC.field(default_factory=list)
    N: Union[Sequence[int], slice] = DC.field(default_factory=list)
    Epoch: Union[Sequence[int], slice] = DC.field(default_factory=list)

def extract_relevant_indices(data: pd.DataFrame, fidelity: Fidelity) -> Union[pd.Index, pd.MultiIndex]:
    """ Given a dataframe in wide-format and a set of fidelity values as an instance of Fidelity, return a
    corresponding index which can be used to select rows corresponding to the given Fidelity instance. """
    # Extract the indices of only the relevant data
    x: pd.DataFrame = data
    x = x.loc[x[fids].isin(DC.asdict(fidelity)).all(axis=1)]
    relevant_index = x.index
    return relevant_index


class LongFormatIndexNames(Enum):
    SAMPLE_IDX = "Sample ID"
    VALUE_TYPE = "Value Type"

class LongFormatOutputTypes(Enum):
    TRUE = "True"
    PREDICTED = "Predicted"

class LongFormatValueTypes(Enum):
    FEATURE = "Feature"
    GROUP = "Group"
    TRUE = LongFormatOutputTypes.TRUE.value
    PREDICTED = LongFormatOutputTypes.PREDICTED.value


def wide_to_long_format(data: pd.DataFrame, output_type: LongFormatOutputTypes) \
        -> pd.DataFrame:
    """ Given a metrics dataframe in wide format, as used by the surrogate, returns an
    equivalent DataFrame in long-format. The parameter `output_type` should be set to one
    of LongFormatOutputTypes.TRUE or LongFormatOutputTypes.PREDICTED to indicate whether
    the values contained within correspond to metrics recorded in the performance
    dataset or were predicted by a surrogate model. The final long-format DataFrame
    treats features and metrics as values, one in each row, and indxes them using a
    combination of the integer index of each row in the wide-format DataFrame, called
    "Sample ID", and the type of value, which can be one of LongFormatValueTypes.FEATURE,
    LongFormatValueTypes.GROUP, LongFormatValueTypes.TRUE or
    LongFormatValueTypes.PREDICTED. LongFormatValueTypes.GROUP corresponds to the values
    of the columns "model_ID" and "fidelity_ID", defining groups of rows belonging to the
    same configuration and fidelity group, respectively.
    """

    _log.debug(f"Attempting to convert input wide-format DataFrame of shape {data.shape} "
               f"to long-format.")
    assert isinstance(output_type, LongFormatOutputTypes), \
        f"The parameter 'output_type' must be a member of the Enum " \
        f"LongFormatOutputTypes, was '{output_type}' of type '{type(output_type)}' " \
        f"instead."

    # Redundant, but ensures semantic compatibility
    output_type = LongFormatValueTypes(output_type.value)

    _log.debug("Splitting wide-format DataFrame's top-level columns into individual "
               "frames.")
    index = data.index.to_frame(index=True, name=LongFormatIndexNames.SAMPLE_IDX.value)

    groups: pd.DataFrame = data.loc[:, "sampling_index"]
    groups: pd.DataFrame = groups.loc[:, ["fidelity_ID", "model_ID"]]
    groups = groups.assign(
        **{LongFormatIndexNames.VALUE_TYPE.value: LongFormatValueTypes.GROUP.value})
    groups = pd.concat([index, groups], axis=1)

    features: pd.DataFrame = data.loc[:, "features"]
    features = features.assign(
        **{LongFormatIndexNames.VALUE_TYPE.value: LongFormatValueTypes.FEATURE.value})
    features = pd.concat([index, features], axis=1)

    labels: pd.DataFrame = data.loc[:, "labels"]
    labels = labels.assign(**{LongFormatIndexNames.VALUE_TYPE.value: output_type.value})
    labels = pd.concat([index, labels], axis=1)

    id_vars = [v.value for v in LongFormatIndexNames]

    _log.debug(f"Performing sanity checks on component frames.")
    frames = {"groups": groups, "features": features, "labels": labels}
    for f1 in frames.items():
        _log.debug(f"LHS frame: {f1[0]}\nColumns: {f1[1].columns}")
        for f2 in frames.items():
            if f1[0] == f2[0]:
                continue
            _log.debug(f"RHS frame: {f2[0]}")
            overlap = f1[1].columns.intersection(f2[1].columns)
            overlap = overlap.difference(id_vars)
            if overlap.size != 0:
                raise RuntimeError(
                    f"Failed to create long-format DataFrame due to the presence of "
                    f"overlapping column names {overlap.tolist()} between '{f1[0]}' and "
                    f"'{f2[0]}'."
                )
    _log.debug("Sanity check successfully passed, melting DataFrame.")

    groups_long = pd.melt(groups, id_vars=id_vars, var_name="variable",
                          value_name="value", ignore_index=True)
    features_long = pd.melt(features, id_vars=id_vars, var_name="variable",
                            value_name="value", ignore_index=True)
    labels_long = pd.melt(labels, id_vars=id_vars, var_name="variable",
                          value_name="value", ignore_index=True)

    long_data = pd.concat([groups_long, features_long, labels_long], axis=0)
    return long_data


def _rank_labels(outputs: pd.DataFrame, axis: int = 0, inplace=False) -> pd.DataFrame:
    """ Given a DataFrame containing the values for any number of outputs in wide-format,
    where all output values are under the top-level column 'labels', return a similar
    DataFrame where each 'label' column's contents are replaced by their respective ranks
    within the original dataframe. """

    x = outputs.copy() if not inplace else outputs
    label_cols = x[["labels"]].columns.tolist()
    ranks = x[label_cols].rank(axis=axis)
    x.loc[:, label_cols] = ranks
    return x


def map_index(index: pd.MultiIndex, mapping: dict, level: int = 0) -> pd.MultiIndex:
    """ Simple functionality that pandas itself really should have. Applies a mapping to only select elements of a
    given MultiIndex and returns the new MultiIndex. """

    index_level_values = index.to_frame(index=False)
    map_fn = lambda y: mapping[y] if y in mapping else y
    index_level_values[level] = list(map(map_fn, index_level_values[level]))
    return pd.MultiIndex.from_frame(index_level_values)


def get_filtered_data(data: pd.DataFrame, relevant_fidelity: Fidelity, relevant_outputs: Sequence[str] = None,
                      adjust_for_minimization: bool = True) -> \
        (pd.DataFrame, Sequence[str]):
    """ Filter out the full dataset according to the given set of fidelities and the
    required outputs. The flag 'adjust_for_minimization', when True (default), adjusts a
    number of metrics such that they are suitable for use in minimization. This includes
    all accuracy metrics, identified by the suffix '-acc' and the metric FLOPS.
    Accuracy metrics are relabeled to '-err' to indicate Error% instead of Accuracy%. """

    _log.debug(f"Cleaning up input data of shape {data.shape}.")

    # Handle the case-sensitivity issues with the "Epoch" parameter
    _log.debug(f"Renaming fidelity 'epoch' to 'Epoch' for consistency.")
    mapping = {"epoch": "Epoch"}
    data.columns = map_index(data.columns, mapping, level=1)

    # Make sure that it is always possible to numerically identify which samples belong
    # together
    _log.debug(f"Generating unique sample IDs for each row.")
    data = data.rename_axis(["Sample ID"], axis=0).reset_index(col_fill="sampling_index",
                                                               col_level=1)
    relevant_index = extract_relevant_indices(data.features, relevant_fidelity)
    data = data.loc[relevant_index]

    rename_accs = {}
    if adjust_for_minimization:
        _log.debug(f"Adjusting metrics values to accomodate minimization.")
        # Convert accuracy to error in order to enable all outputs to have the same
        # ordering i.e. a smaller value is better
        acc_cols = [c for c in data.columns if "-acc" in c[1]]
        data.loc[:, acc_cols] = data.loc[:, acc_cols].rsub(100).values
        rename_accs = {c[1]: c[1].replace("-acc", "-err") for c in acc_cols}

        flops_cols = [c for c in data.columns if "FLOPS" in c[1]]
        data.loc[:, flops_cols] = data.loc[:, flops_cols].mul(-1).values

    # Re-scale size_MB to size_KB in order to make plotting easier
    size_cols = [c for c in data.columns if "size_MB" in c[1]]
    data.loc[:, size_cols] = data.loc[:, size_cols].mul(1000).values
    rename_sizes = {c[1]: c[1].replace("_MB", "_KB") for c in size_cols}

    # Filter out the outputs if needed
    if relevant_outputs is not None:
        _log.debug(f"Restricting the chosen set of labels to {relevant_outputs}.")
        drop_cols = data.labels.columns.difference(relevant_outputs)
        data = data.drop([("labels", c) for c in drop_cols], axis=1)

    _log.debug(f"Renaming metrics, if needed.")
    rename_map = {**rename_accs, **rename_sizes}
    data.columns = map_index(data.columns, rename_map, level=1)

    _log.debug(f"Finished cleaning up input data of shape {data.shape}.")
    return data


def get_correlations(data: pd.DataFrame, inplace: bool = False) -> pd.DataFrame:
    """ Given a wide-format DataFrame, rank the outputs specified under the top-level
    column 'labels' across all the given indices and return a DataFrame containing the
    corresponding Kendall's tau rank correlation values and p-values, with the relevant
    "labels" serving as both index and column values. The top-level column values are
    "correlation" and "p-value". This DataFrame will be of shape [l, 2*l], where
    'l' is the number of columns under 'labels'. """

    _log.debug(f"Generating ranks.")
    ranks = _rank_labels(data, inplace=inplace)
    index = ranks.labels.columns.copy()

    _log.debug(f"Generating correlations.")
    corrdf = pd.DataFrame(None, index=index, columns=index)
    pvaldf = pd.DataFrame(None, index=index, columns=index)

    # Calculate Kendall's tau value for rank correlations
    combinations = index.size ** 2
    i = 0
    for label1 in index.copy():
        for label2 in index.copy():
            i += 1
            _log.debug(f"Generating correlations for ({label1}, {label2}) "
                       f"({i}/{combinations})")
            vals1 = ranks[("labels", label1)].values.squeeze()
            vals2 = ranks[("labels", label2)].values.squeeze()
            corr, pval = kendalltau(vals1, vals2, nan_policy="omit")
            corrdf.loc[label1, label2] = corr
            pvaldf.loc[label1, label2] = pval

    correlations = pd.concat({"correlations": corrdf, "p-values": pvaldf}, axis=1)

    _log.debug(f"Correlations DataFrame has shape: {correlations.shape}")
    return correlations


# TODO: Re-define this in order to change the default arguments, e.g. to change 'multioutput'
score_funcs = {
    "R2": partial(r2_score, multioutput="raw_values"),
    "RMSE": partial(mean_squared_error, multioutput="raw_values", squared=False),
    "MAPE": partial(mean_absolute_percentage_error, multioutput="raw_values"),
    "MedAE": partial(median_absolute_error, multioutput="raw_values"),
}


def get_scores(data: pd.DataFrame, group_cols: Sequence[str] = None):
    """ Given a long-format dataframe of benchmark data, with all value variables under the column 'output_var',
    the label for "True" vs "Predicted" values under the column 'output_type' and all output values under the
    column 'value', return a long-format DataFrame with the regression analysis scores of the given dataset
    such that the column 'value' now refers to a metric's score, the metric name is stored in the column
    'metric_type' and the other columns serve as identifier variables corresponding to the column 'output_var'
    from the original data and any other grouping columns specified in 'group_cols'. """

    group_cols = [] if group_cols is None else group_cols
    x = data.set_index(data.columns.difference(["value"]).to_list())
    g = x.unstack(["output_type"]).droplevel(0, axis=1).groupby(["output_var"] + group_cols)

    scores = {k: g.apply(lambda x: f(x.loc[:, "True"], x.loc[:, "Predicted"])) for k, f in score_funcs.items()}
    scores = pd.DataFrame(scores, columns=pd.Index(list(scores.keys()), name="metric_type"))
    scores = scores.explode(list(scores.keys()))
    scores = scores.stack(["metric_type"]).rename("value").reset_index()
    return scores

    # TODO: Test for multiple outputs


def get_test_frac_data(test_frac: float, filter_loss: bool = True):
    """ Load up the data for the model trained using a particular dataset split, specified by the float
    'test_frac', from disk, filter and clean the data and return it as a wide-format dataframe with
    2-level MultiIndex columns. """
    datadir = basedir / f"test-frac-{test_frac}"
    test_set_pth = datadir / "test_set.pkl.gz"
    test_pred_pth = datadir / "test_pred.pkl.gz"
    test_set = pd.read_pickle(test_set_pth)
    ypred = pd.read_pickle(test_pred_pth)
    test_set.rename({"labels": "true"}, axis=1, inplace=True)
    test_set.rename({"epoch": "Epoch"}, axis=1, inplace=True)
    test_set.drop([c for c in test_set.columns if "-loss" in c[1]], axis=1, inplace=True)
    ypred.drop([c for c in ypred.columns if "-loss" in c[1]], axis=1, inplace=True)
    ypred = ypred.droplevel(0, axis=1)
    ypred.columns = pd.MultiIndex.from_product([["predicted"], ypred.columns.to_list()])
    assert ypred.shape[0] == test_set.shape[0]
    ypred.index = test_set.index
    data = pd.concat([test_set, ypred], axis=1)
    return data

def get_filtered_corr(data: pd.DataFrame, extra_group_by: Sequence[str] = None):
    correlations = get_correlations(data=data, extra_group_by=extra_group_by)
    x = correlations
    x = x.loc[x.output_type=="Predicted"]
    x = x.loc[x["val-output_type"]=="True"]
    x = x.loc[x["Metric Name"]=="Kendall-Tau"]
    return x[["Value Type", "value"]]

# maxfidelity = Fidelity(N=[5], W=[16], Resolution=[1.0], Epoch=[200])
# maxcapacity = Fidelity(N=[5], W=[16], Resolution=[1.0], Epoch=list(range(1, 201)))
# maxbudget = Fidelity(N=[1, 3, 5], W=[4, 8, 16], Resolution=[0.25, 0.5, 1.0], Epoch=[200])
fids = ["Resolution", "W", "N", "Epoch"]
